from pathlib import Path
from typing import Any, Optional, Type, Union, Dict

from pydantic import BaseModel, validator

from phidata.utils.log import logger


class InfraResource(BaseModel):
    """Base class for all Phidata infrastructure resources.
    All Models in the phidata.infra.*.resource modules are expected to be subclasses of this Model.

    We use a pydantic model for resources because the data which creates the resource
    may come from an external sources like users or an api.
    This data needs to be validated & type checked for which we use pydantic.
    """

    # name of resource
    name: Optional[str] = None
    # type of resource
    resource_type: Optional[str] = None

    # resource management
    # If True, skip resource creation if an active resources with the same name exists.
    use_cache: bool = True
    # Force the function to be implemented
    # i.e. force recreate, force delete, force update
    force: bool = False

    # If enabled=False: mark skip_create, skip_delete, skip_update as True
    enabled: bool = True
    # If True, phi does not create the resource
    skip_create: bool = False
    # If True, phi does not read the resource
    skip_read: bool = False
    # If True, phi does not update the resource
    skip_update: bool = False
    # If True, phi does not delete the resource
    skip_delete: bool = False

    # If True, waits for the resource to be created
    wait_for_creation: bool = True
    # If True, waits for the resource to be updated
    wait_for_update: bool = True
    # If True, waits for the resource to be deleted
    wait_for_deletion: bool = True
    # The amount of time in seconds to wait between attempts.
    waiter_delay: int = 30
    # The maximum number of attempts to be made.
    waiter_max_attempts: int = 50

    active_resource: Optional[Any] = None
    active_resource_class: Optional[Type] = None

    resource_file: Optional[Union[str, Path]] = None

    def get_resource_name(self) -> Optional[str]:
        return self.name

    def get_resource_type(self) -> Optional[str]:
        return self.resource_type

    @validator("force", pre=True, always=True)
    def set_force(cls, force):
        from os import getenv

        force = force or getenv("PHI_WS_FORCE", False)
        return force

    @validator("skip_create", pre=True, always=True)
    def set_skip_create(cls, skip_create, values):
        skip_resource = not values.get("enabled", True)
        return True if skip_resource else skip_create

    @validator("skip_update", pre=True, always=True)
    def set_skip_update(cls, skip_update, values):
        skip_resource = not values.get("enabled", True)
        return True if skip_resource else skip_update

    @validator("skip_delete", pre=True, always=True)
    def set_skip_delete(cls, skip_delete, values):
        skip_resource = not values.get("enabled", True)
        return True if skip_resource else skip_delete

    """
    ## Functions to be implemented by subclasses
    def create(self, infra_api_client: InfraApiClient) -> bool:
    def read(self, infra_api_client: InfraApiClient) -> bool:
    def update(self, infra_api_client: InfraApiClient) -> bool:
    def delete(self, infra_api_client: InfraApiClient) -> bool:
    def is_active(self, infra_api_client: InfraApiClient) -> bool:
    """

    class Config:
        # https://pydantic-docs.helpmanual.io/usage/model_config/
        # If we need to use an alias for fields of subclasses, eg: Kubeconfig
        allow_population_by_field_name = True
        arbitrary_types_allowed = True
        use_enum_values = True

    @staticmethod
    def get_workspace_config_file_path() -> Optional[Path]:
        from os import getenv
        from phidata.constants import WORKSPACE_CONFIG_FILE_ENV_VAR

        workspace_config_file_path = getenv(WORKSPACE_CONFIG_FILE_ENV_VAR, None)
        if workspace_config_file_path is not None:
            return Path(workspace_config_file_path)
        return None

    @staticmethod
    def get_workspace_root_path() -> Optional[Path]:
        from os import getenv
        from phidata.constants import WORKSPACE_ROOT_ENV_VAR

        workspace_root_path = getenv(WORKSPACE_ROOT_ENV_VAR, None)
        if workspace_root_path is not None:
            return Path(workspace_root_path)
        return None

    @staticmethod
    def get_workspace_config_dir() -> Optional[Path]:
        from os import getenv
        from phidata.constants import WORKSPACE_CONFIG_DIR_ENV_VAR

        workspace_config_dir = getenv(WORKSPACE_CONFIG_DIR_ENV_VAR, None)
        if workspace_config_dir is not None:
            return Path(workspace_config_dir)
        return None

    @staticmethod
    def get_workflows_dir() -> Optional[Path]:
        from os import getenv
        from phidata.constants import WORKFLOWS_DIR_ENV_VAR

        workflows_dir = getenv(WORKFLOWS_DIR_ENV_VAR, None)
        if workflows_dir is not None:
            return Path(workflows_dir)
        return None

    @staticmethod
    def get_meta_dir() -> Optional[Path]:
        from os import getenv
        from phidata.constants import META_DIR_ENV_VAR

        meta_dir = getenv(META_DIR_ENV_VAR, None)
        if meta_dir is not None:
            return Path(meta_dir)
        return None

    @staticmethod
    def get_notebooks_dir() -> Optional[Path]:
        from os import getenv
        from phidata.constants import NOTEBOOKS_DIR_ENV_VAR

        notebooks_dir = getenv(NOTEBOOKS_DIR_ENV_VAR, None)
        if notebooks_dir is not None:
            return Path(notebooks_dir)
        return None

    @staticmethod
    def get_products_dir() -> Optional[Path]:
        from os import getenv
        from phidata.constants import PRODUCTS_DIR_ENV_VAR

        products_dir = getenv(PRODUCTS_DIR_ENV_VAR, None)
        if products_dir is not None:
            return Path(products_dir)
        return None

    @staticmethod
    def get_scripts_dir() -> Optional[Path]:
        from os import getenv
        from phidata.constants import SCRIPTS_DIR_ENV_VAR

        scripts_dir = getenv(SCRIPTS_DIR_ENV_VAR, None)
        if scripts_dir is not None:
            return Path(scripts_dir)
        return None

    @staticmethod
    def get_storage_dir() -> Optional[Path]:
        from os import getenv
        from phidata.constants import STORAGE_DIR_ENV_VAR

        storage_dir = getenv(STORAGE_DIR_ENV_VAR, None)
        if storage_dir is not None:
            return Path(storage_dir)
        return None

    def read_yaml_file(self, file_path: Optional[Path]) -> Optional[Dict[str, Any]]:
        if file_path is not None and file_path.exists() and file_path.is_file():
            import yaml

            # logger.debug(f"Reading {file_path}")
            data_from_file = yaml.safe_load(file_path.read_text())
            if data_from_file is not None and isinstance(data_from_file, dict):
                return data_from_file
            else:
                logger.error(f"Invalid file: {file_path}")
        return None

    def save_resource_file(self) -> bool:

        if self.resource_file is not None:
            resource_file_path: Optional[Path] = None
            if isinstance(self.resource_file, str):
                resource_file_path = Path(self.resource_file)
            elif isinstance(self.resource_file, Path):
                resource_file_path = self.resource_file

            if resource_file_path is None or not isinstance(resource_file_path, Path):
                logger.error(f"Invalid resource_file: {resource_file_path}")

            try:
                if not resource_file_path.exists():
                    resource_file_path.parent.mkdir(parents=True, exist_ok=True)
                    resource_file_path.touch(exist_ok=True)
                resource_file_path.write_text(self.json(indent=2))
                logger.debug(f"Resource stored at: {str(resource_file_path)}")
                return True
            except Exception as e:
                logger.error("Could not write resource to file")
                logger.error(e)
        return False



from typing import Optional, Any, Dict, Literal

from phidata.infra.aws.api_client import AwsApiClient
from phidata.infra.aws.resource.base import AwsResource
from phidata.utils.cli_console import print_info, print_error
from phidata.utils.log import logger


class EksAddon(AwsResource):
    """
    # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/eks.html
    """

    resource_type = "EksAddon"
    service_name = "eks"

    # Addon name
    name: str
    # EKS cluster name
    cluster_name: str
    # Addon version
    version: Optional[str] = None

    service_account_role_arn: Optional[str] = None
    resolve_conflicts: Optional[Literal["OVERWRITE", "NONE", "PRESERVE"]] = None
    client_request_token: Optional[str] = None
    tags: Optional[Dict[str, str]] = None

    preserve: Optional[bool] = False

    # provided by api on create
    created_at: Optional[str] = None
    status: Optional[str] = None

    wait_for_creation = False
    wait_for_deletion = False
    wait_for_update = False

    def _create(self, aws_client: AwsApiClient) -> bool:
        """Creates the EksAddon

        Args:
            aws_client: The AwsApiClient for the current cluster
        """
        print_info(f"Creating {self.get_resource_type()}: {self.get_resource_name()}")

        # create a dict of args which are not null, otherwise aws type validation fails
        not_null_args: Dict[str, Any] = {}
        if self.version:
            not_null_args["addonVersion"] = self.version
        if self.service_account_role_arn:
            not_null_args["serviceAccountRoleArn"] = self.service_account_role_arn
        if self.resolve_conflicts:
            not_null_args["resolveConflicts"] = self.resolve_conflicts
        if self.client_request_token:
            not_null_args["clientRequestToken"] = self.client_request_token
        if self.tags:
            not_null_args["tags"] = self.tags

        # Step 1: Create EksAddon
        service_client = self.get_service_client(aws_client)
        try:
            create_response = service_client.create_addon(
                clusterName=self.cluster_name,
                addonName=self.name,
                **not_null_args,
            )
            logger.debug(f"EksAddon: {create_response}")
            # logger.debug(f"EksAddon type: {type(create_response)}")

            # Validate Cluster creation
            self.created_at = create_response.get("addon", {}).get("createdAt", None)
            self.status = create_response.get("addon", {}).get("status", None)
            logger.debug(f"created_at: {self.created_at}")
            logger.debug(f"status: {self.status}")
            if self.created_at is not None:
                print_info(f"EksAddon created: {self.name}")
                self.active_resource = create_response
                return True
        except service_client.exceptions.ResourceInUseException:
            print_info(f"Addon already exists: {self.name}")
            return True
        except Exception as e:
            print_error(f"{self.get_resource_type()} could not be created.")
            print_error(e)
        return False

    def post_create(self, aws_client: AwsApiClient) -> bool:

        # Wait for Addon to be created
        if self.wait_for_creation:
            try:
                print_info(f"Waiting for {self.get_resource_type()} to be active.")
                waiter = self.get_service_client(aws_client).get_waiter("addon_active")
                waiter.wait(
                    clusterName=self.cluster_name,
                    addonName=self.name,
                    WaiterConfig={
                        "Delay": self.waiter_delay,
                        "MaxAttempts": self.waiter_max_attempts,
                    },
                )
            except Exception as awe:
                # print_error(f"Waiter failed: {awe}")
                pass
        return True

    def _read(self, aws_client: AwsApiClient) -> Optional[Any]:
        """Returns the EksAddon

        Args:
            aws_client: The AwsApiClient for the current cluster
        """
        from botocore.exceptions import ClientError

        logger.debug(f"Reading {self.get_resource_type()}: {self.get_resource_name()}")

        service_client = self.get_service_client(aws_client)
        try:
            describe_response = service_client.describe_addon(
                clusterName=self.cluster_name, addonName=self.name
            )
            # logger.debug(f"EksAddon: {describe_response}")
            # logger.debug(f"EksAddon type: {type(describe_response)}")
            addon_dict = describe_response.get("addon", {})

            self.created_at = addon_dict.get("createdAt", None)
            self.status = addon_dict.get("status", None)
            logger.debug(f"EksAddon created_at: {self.created_at}")
            logger.debug(f"EksAddon status: {self.status}")
            if self.created_at is not None:
                logger.debug(f"EksAddon found: {self.name}")
                self.active_resource = describe_response
        except ClientError as ce:
            logger.debug(f"ClientError: {ce}")
        except Exception as e:
            print_error(f"Error reading {self.get_resource_type()}.")
            print_error(e)
        return self.active_resource

    def _delete(self, aws_client: AwsApiClient) -> bool:
        """Deletes the EksAddon

        Args:
            aws_client: The AwsApiClient for the current cluster
        """
        print_info(f"Deleting {self.get_resource_type()}: {self.get_resource_name()}")

        # create a dict of args which are not null, otherwise aws type validation fails
        not_null_args: Dict[str, Any] = {}
        if self.preserve:
            not_null_args["preserve"] = self.preserve

        # Step 1: Delete EksAddon
        service_client = self.get_service_client(aws_client)
        self.active_resource = None
        try:
            delete_response = service_client.delete_addon(
                clusterName=self.cluster_name, addonName=self.name, **not_null_args
            )
            logger.debug(f"EksAddon: {delete_response}")
            # logger.debug(f"EksAddon type: {type(delete_response)}")
            print_info(
                f"{self.get_resource_type()}: {self.get_resource_name()} deleted"
            )
            return True
        except Exception as e:
            print_error(f"{self.get_resource_type()} could not be deleted.")
            print_error("Please try again or delete resources manually.")
            print_error(e)
        return False

    def post_delete(self, aws_client: AwsApiClient) -> bool:

        # Wait for Addon to be deleted
        if self.wait_for_deletion:
            try:
                print_info(f"Waiting for {self.get_resource_type()} to be deleted.")
                waiter = self.get_service_client(aws_client).get_waiter("addon_deleted")
                waiter.wait(
                    clusterName=self.cluster_name,
                    addonName=self.name,
                    WaiterConfig={
                        "Delay": self.waiter_delay,
                        "MaxAttempts": self.waiter_max_attempts,
                    },
                )
            except Exception as awe:
                print_error(f"Waiter failed: {awe}")
        return True



from pathlib import Path
from typing import Optional

import kubernetes

from phidata.infra.base.api_client import InfraApiClient
from phidata.infra.k8s.exceptions import K8sApiClientException
from phidata.utils.cli_console import print_error
from phidata.utils.log import logger


class K8sApiClient(InfraApiClient):
    """
    This class is a wrapper around the kubernetes clients to use with a K8sWorker
    """

    def __init__(
        self,
        context: Optional[str] = None,
        kubeconfig_path: Optional[Path] = None,
    ):

        super().__init__()
        # logger.debug(f"Creating K8sApiClient")
        # Cluster configuration
        self.context: Optional[str] = context
        self.kubeconfig_path: Optional[Path] = kubeconfig_path
        self.configuration: Optional[kubernetes.client.Configuration] = None

        # kubernetes API clients
        self._api_client: Optional[kubernetes.client.ApiClient] = None
        self._apps_v1_api: Optional[kubernetes.client.AppsV1Api] = None
        self._core_v1_api: Optional[kubernetes.client.CoreV1Api] = None
        self._rbac_auth_v1_api: Optional[
            kubernetes.client.RbacAuthorizationV1Api
        ] = None
        self._storage_v1_api: Optional[kubernetes.client.StorageV1Api] = None
        self._apiextensions_v1_api: Optional[
            kubernetes.client.ApiextensionsV1Api
        ] = None
        self._custom_objects_api: Optional[kubernetes.client.CustomObjectsApi] = None

        logger.debug(f"**-+-** K8sApiClient created")

    def is_initialized(self) -> bool:
        if self.api_client is not None:
            return True
        return False

    def create_api_client(self) -> kubernetes.client.ApiClient:
        """Create a kubernetes.client.ApiClient"""

        logger.debug("Creating kubernetes.client.ApiClient")
        api_client = None
        try:
            self.configuration = kubernetes.client.Configuration()
            try:
                kubernetes.config.load_kube_config(
                    client_configuration=self.configuration, context=self.context
                )
            except kubernetes.config.ConfigException as config_exc:
                # Usually because the context is not in the kubeconfig
                kubernetes.config.load_kube_config(
                    client_configuration=self.configuration
                )
            logger.debug(f"\thost: {self.configuration.host}")
            api_client = kubernetes.client.ApiClient(self.configuration)
            # logger.debug(f"\tApiClient: {api_client}")
        except Exception as e:
            print_error(e)
        if api_client is None:
            raise K8sApiClientException("Failed creating Kubernetes ApiClient")
        return api_client

    ######################################################
    ## K8s APIs as properties
    # APIs are lazy-initialized when needed
    ######################################################

    @property
    def api_client(self) -> kubernetes.client.ApiClient:
        if self._api_client is None:
            self._api_client = self.create_api_client()
        return self._api_client

    @property
    def apps_v1_api(self) -> kubernetes.client.AppsV1Api:
        if self._apps_v1_api is None:
            self._apps_v1_api = kubernetes.client.AppsV1Api(self.api_client)
        return self._apps_v1_api

    @property
    def core_v1_api(self) -> kubernetes.client.CoreV1Api:
        if self._core_v1_api is None:
            self._core_v1_api = kubernetes.client.CoreV1Api(self.api_client)
        return self._core_v1_api

    @property
    def rbac_auth_v1_api(self) -> kubernetes.client.RbacAuthorizationV1Api:
        if self._rbac_auth_v1_api is None:
            self._rbac_auth_v1_api = kubernetes.client.RbacAuthorizationV1Api(
                self.api_client
            )
        return self._rbac_auth_v1_api

    @property
    def storage_v1_api(self) -> kubernetes.client.StorageV1Api:
        if self._storage_v1_api is None:
            self._storage_v1_api = kubernetes.client.StorageV1Api(self.api_client)
        return self._storage_v1_api

    @property
    def apiextensions_v1_api(self) -> kubernetes.client.ApiextensionsV1Api:
        if self._apiextensions_v1_api is None:
            self._apiextensions_v1_api = kubernetes.client.ApiextensionsV1Api(
                self.api_client
            )
        return self._apiextensions_v1_api

    @property
    def custom_objects_api(self) -> kubernetes.client.CustomObjectsApi:
        if self._custom_objects_api is None:
            self._custom_objects_api = kubernetes.client.CustomObjectsApi(
                self.api_client
            )
        return self._custom_objects_api



from typing import Optional, Literal, List, Tuple, Dict, Any, Union

from phidata.asset.aws import AwsAsset, AwsAssetArgs
from phidata.task import Task
from phidata.check import Check
from phidata.infra.aws.resource.s3.bucket import S3Bucket

from phidata.utils.enums import ExtendedEnum
from phidata.utils.log import logger


class S3DatasetType(ExtendedEnum):
    CSV = "CSV"
    JSON = "JSON"
    PARQUET = "PARQUET"


class S3DatasetBaseArgs(AwsAssetArgs):
    # Dataset Name. Can be provided instead of "table" argument
    name: Optional[str] = None

    # Glue/Athena catalog: Table name.
    table: Optional[str] = None
    # Glue/Athena catalog: Database name.
    database: Optional[str] = None
    # Type of dataset: CSV, JSON, PARQUET
    dataset_type: Optional[S3DatasetType] = None
    # The type of the Glue Table. Set to EXTERNAL_TABLE if None.
    table_type: Optional[str] = None
    # Glue/Athena catalog: Table description
    table_description: Optional[str] = None
    # write mode – append (Default), overwrite, overwrite_partitions.
    write_mode: Optional[Literal["append", "overwrite", "overwrite_partitions"]] = None

    # S3 path for the dataset. If not provided, inferred from
    # path_prefix/top_level_dir/database/table or
    # s3://bucket/top_level_dir/database/table
    path: Optional[str] = None
    # Top level directory for all s3 datasets.
    top_level_dir: Optional[str] = "datasets"
    # Prefix for the dataset path. Used to infer dataset path if not provided as
    # path_prefix/top_level_dir/database/table
    path_prefix: Optional[str] = None
    # S3 bucket for the dataset. Used to infer dataset path if not provided as
    # s3://bucket/top_level_dir/database/table
    bucket: Optional[S3Bucket] = None
    # If dataset=True, add a filename prefix to the output files.
    filename_prefix: Optional[str] = None

    # Dictionary of columns names and Athena/Glue types to be casted.
    # Useful when you have columns with undetermined or mixed data types.
    # (e.g. {‘col name’: ‘bigint’, ‘col2 name’: ‘int’})
    dtype: Optional[Dict[str, str]] = None
    # Glue/Athena catalog: Key/value pairs to tag the table.
    parameters: Optional[Dict[str, str]] = None
    # Glue/Athena catalog: Columns names and the related comments
    # (e.g. {‘col0’: ‘Column 0.’, ‘col1’: ‘Column 1.’, ‘col2’: ‘Partition.’}).
    columns_comments: Optional[Dict[str, str]] = None

    # List of column names that will be used to create partitions.
    partition_cols: Optional[List[str]] = None
    # Tuple consisting of the column names used for
    # bucketing as the first element and the number of buckets as the second element.
    # Only str, int and bool are supported as column data types for bucketing.
    bucketing_info: Optional[Tuple[List[str], int]] = None
    # If True increases parallelism level during the partitions writing.
    # It will decrease the writing time and increase the memory usage.
    # https://aws-sdk-pandas.readthedocs.io/en/2.17.0/tutorials/022%20-%20Writing%20Partitions%20Concurrently.html
    concurrent_partitioning: Optional[bool] = None

    # If True and mode="overwrite", creates an archived version of the table catalog before updating it.
    catalog_versioning: Optional[bool] = None
    # If True allows schema evolution (new or missing columns), otherwise a exception will be raised.
    # True by default. (Only considered if mode in (“append”, “overwrite_partitions”))
    schema_evolution: Optional[bool] = None
    # The ID of the Data Catalog from which to retrieve Databases.
    # If none is provided, the AWS account ID is used by default.
    catalog_id: Optional[str] = None

    # True to enable concurrent requests, False to disable multiple threads.
    # If enabled os.cpu_count() will be used as the max number of threads.
    # If integer is provided, specified number is used.
    use_threads: Optional[Union[bool, int]] = None
    # Boto3 Session. The default boto3 Session will be used if boto3_session receive None.
    boto3_session: Optional[Any] = None
    # Forwarded to botocore requests.
    # e.g. s3_additional_kwargs={‘ServerSideEncryption’: ‘aws:kms’, ‘SSEKMSKeyId’: ‘YOUR_KMS_KEY_ARN’}
    s3_additional_kwargs: Optional[Dict[str, Any]] = None

    # Note: These arguments are not exposed using properties.
    # But they are included in args incase we want to pass down these values to aws wrangler.
    # regular_partitions (bool) – Create regular partitions (Non projected partitions) on Glue Catalog.
    # Disable when you will work only with Partition Projection.
    regular_partitions: Optional[bool] = None
    # Enable Partition Projection on Athena
    # https://docs.aws.amazon.com/athena/latest/ug/partition-projection.html
    projection_enabled: Optional[bool] = None
    # Dictionary of partitions names and Athena projections types. Valid types: “enum”, “integer”, “date”, “injected”
    # https://docs.aws.amazon.com/athena/latest/ug/partition-projection-supported-types.html
    # e.g. {‘col_name’: ‘enum’, ‘col2_name’: ‘integer’}
    projection_types: Optional[Dict[str, str]] = None
    # Dictionary of partitions names and Athena projections ranges.
    # https://docs.aws.amazon.com/athena/latest/ug/partition-projection-supported-types.html
    # e.g. {‘col_name’: ‘0,10’, ‘col2_name’: ‘-1,8675309’}
    projection_ranges: Optional[Dict[str, str]] = None
    # Dictionary of partitions names and Athena projections values.
    # https://docs.aws.amazon.com/athena/latest/ug/partition-projection-supported-types.html
    # e.g. {‘col_name’: ‘A,B,Unknown’, ‘col2_name’: ‘foo,boo,bar’}
    projection_values: Optional[Dict[str, str]] = None
    # Dictionary of partitions names and Athena projections intervals.
    # https://docs.aws.amazon.com/athena/latest/ug/partition-projection-supported-types.html
    # e.g. {‘col_name’: ‘1’, ‘col2_name’: ‘5’}
    projection_intervals: Optional[Dict[str, str]] = None
    # Dictionary of partitions names and Athena projections digits.
    # https://docs.aws.amazon.com/athena/latest/ug/partition-projection-supported-types.html
    # e.g. {‘col_name’: ‘1’, ‘col2_name’: ‘2’}
    projection_digits: Optional[Dict[str, str]] = None
    # Dictionary of partitions names and Athena projections formats.
    # https://docs.aws.amazon.com/athena/latest/ug/partition-projection-supported-types.html
    # e.g. {‘col_date’: ‘yyyy-MM-dd’, ‘col2_timestamp’: ‘yyyy-MM-dd HH:mm:ss’}
    projection_formats: Optional[Dict[str, str]] = None
    # Value which is allows Athena to properly map partition values if the S3 file locations
    # do not follow a typical …/column=value/… pattern.
    # https://docs.aws.amazon.com/athena/latest/ug/partition-projection-setting-up.html
    # e.g. s3://bucket/table_root/a=${a}/${b}/some_static_subdirectory/${c}/
    projection_storage_location_template: Optional[str] = None


class S3DatasetBase(AwsAsset):
    def __init__(self) -> None:
        super().__init__()
        self.args: S3DatasetBaseArgs = S3DatasetBaseArgs()

    @property
    def table(self) -> Optional[str]:
        return self.args.table

    @table.setter
    def table(self, table: str) -> None:
        if table is not None:
            self.args.table = table

    @property
    def database(self) -> Optional[str]:
        return self.args.database

    @database.setter
    def database(self, database: str) -> None:
        if database is not None:
            self.args.database = database

    @property
    def dataset_type(self) -> Optional[S3DatasetType]:
        return self.args.dataset_type

    @dataset_type.setter
    def dataset_type(self, dataset_type: S3DatasetType) -> None:
        if dataset_type is not None:
            self.args.dataset_type = dataset_type

    @property
    def table_type(self) -> Optional[str]:
        return self.args.table_type

    @table_type.setter
    def table_type(self, table_type: str) -> None:
        if table_type is not None:
            self.args.table_type = table_type

    @property
    def table_description(self) -> Optional[str]:
        return self.args.table_description

    @table_description.setter
    def table_description(self, table_description: str) -> None:
        if table_description is not None:
            self.args.table_description = table_description

    @property
    def write_mode(
        self,
    ) -> Optional[Literal["append", "overwrite", "overwrite_partitions"]]:
        return self.args.write_mode

    @write_mode.setter
    def write_mode(
        self, write_mode: Literal["append", "overwrite", "overwrite_partitions"]
    ) -> None:
        if write_mode is not None:
            self.args.write_mode = write_mode

    @property
    def path(self) -> Optional[str]:
        return self.args.path

    @path.setter
    def path(self, path: str) -> None:
        if path is not None:
            self.args.path = path

    @property
    def top_level_dir(self) -> Optional[str]:
        return self.args.top_level_dir

    @top_level_dir.setter
    def top_level_dir(self, top_level_dir: str) -> None:
        if top_level_dir is not None:
            self.args.top_level_dir = top_level_dir

    @property
    def path_prefix(self) -> Optional[str]:
        return self.args.path_prefix

    @path_prefix.setter
    def path_prefix(self, path_prefix: str) -> None:
        if path_prefix is not None:
            self.args.path_prefix = path_prefix

    @property
    def bucket(self) -> Optional[S3Bucket]:
        return self.args.bucket

    @bucket.setter
    def bucket(self, bucket: S3Bucket) -> None:
        if bucket is not None:
            self.args.bucket = bucket

    @property
    def filename_prefix(self) -> Optional[str]:
        return self.args.filename_prefix

    @filename_prefix.setter
    def filename_prefix(self, filename_prefix: str) -> None:
        if filename_prefix is not None:
            self.args.filename_prefix = filename_prefix

    @property
    def dtype(self) -> Optional[Dict[str, str]]:
        return self.args.dtype

    @dtype.setter
    def dtype(self, dtype: Dict[str, str]) -> None:
        if dtype is not None:
            self.args.dtype = dtype

    @property
    def parameters(self) -> Optional[Dict[str, str]]:
        return self.args.parameters

    @parameters.setter
    def parameters(self, parameters: Dict[str, str]) -> None:
        if parameters is not None:
            self.args.parameters = parameters

    @property
    def columns_comments(self) -> Optional[Dict[str, str]]:
        return self.args.columns_comments

    @columns_comments.setter
    def columns_comments(self, columns_comments: Dict[str, str]) -> None:
        if columns_comments is not None:
            self.args.columns_comments = columns_comments

    @property
    def partition_cols(self) -> Optional[List[str]]:
        return self.args.partition_cols

    @partition_cols.setter
    def partition_cols(self, partition_cols: List[str]) -> None:
        if partition_cols is not None:
            self.args.partition_cols = partition_cols

    @property
    def bucketing_info(self) -> Optional[Tuple[List[str], int]]:
        return self.args.bucketing_info

    @bucketing_info.setter
    def bucketing_info(self, bucketing_info: Tuple[List[str], int]) -> None:
        if bucketing_info is not None:
            self.args.bucketing_info = bucketing_info

    @property
    def concurrent_partitioning(self) -> Optional[bool]:
        return self.args.concurrent_partitioning

    @concurrent_partitioning.setter
    def concurrent_partitioning(self, concurrent_partitioning: bool) -> None:
        if concurrent_partitioning is not None:
            self.args.concurrent_partitioning = concurrent_partitioning

    @property
    def catalog_versioning(self) -> Optional[bool]:
        return self.args.catalog_versioning

    @catalog_versioning.setter
    def catalog_versioning(self, catalog_versioning: bool) -> None:
        if catalog_versioning is not None:
            self.args.catalog_versioning = catalog_versioning

    @property
    def schema_evolution(self) -> Optional[bool]:
        return self.args.schema_evolution

    @schema_evolution.setter
    def schema_evolution(self, schema_evolution: bool) -> None:
        if schema_evolution is not None:
            self.args.schema_evolution = schema_evolution

    @property
    def catalog_id(self) -> Optional[str]:
        return self.args.catalog_id

    @catalog_id.setter
    def catalog_id(self, catalog_id: str) -> None:
        if catalog_id is not None:
            self.args.catalog_id = catalog_id

    @property
    def use_threads(self) -> Optional[Union[bool, int]]:
        return self.args.use_threads

    @use_threads.setter
    def use_threads(self, use_threads: Union[bool, int]) -> None:
        if use_threads is not None:
            self.args.use_threads = use_threads

    @property
    def boto3_session(self) -> Optional[Any]:
        return self.args.boto3_session

    @boto3_session.setter
    def boto3_session(self, boto3_session: Any) -> None:
        if boto3_session is not None:
            self.args.boto3_session = boto3_session

    @property
    def s3_additional_kwargs(self) -> Optional[Dict[str, Any]]:
        return self.args.s3_additional_kwargs

    @s3_additional_kwargs.setter
    def s3_additional_kwargs(self, s3_additional_kwargs: Dict[str, Any]) -> None:
        if s3_additional_kwargs is not None:
            self.args.s3_additional_kwargs = s3_additional_kwargs

    @property
    def uri(self) -> str:
        if self.path is not None:
            return self.path
        else:
            if self.path_prefix is not None:
                return "{}/{}{}/{}".format(
                    self.path_prefix,
                    f"{self.top_level_dir}/" if self.top_level_dir else "",
                    self.database,
                    self.table,
                )
            if self.bucket is not None:
                return "{}/{}{}/{}".format(
                    self.bucket.get_uri(),
                    f"{self.top_level_dir}/" if self.top_level_dir else "",
                    self.database,
                    self.table,
                )
        return ""



from typing import Dict, List, Optional

from pydantic import BaseModel

from phidata.infra.k8s.enums.api_version import ApiVersion
from phidata.infra.k8s.enums.kind import Kind
from phidata.infra.k8s.resource.core.v1.service_account import (
    ServiceAccount,
    LocalObjectReference,
    ObjectReference,
)
from phidata.infra.k8s.create.common.labels import create_component_labels_dict
from phidata.infra.k8s.resource.meta.v1.object_meta import ObjectMeta
from phidata.utils.common import get_default_sa_name
from phidata.utils.log import logger


class CreateServiceAccount(BaseModel):
    sa_name: str
    app_name: str
    automount_service_account_token: Optional[bool] = None
    image_pull_secrets: Optional[List[str]] = None
    secrets: Optional[List[ObjectReference]] = None
    namespace: Optional[str] = None
    labels: Optional[Dict[str, str]] = None

    def create(self) -> Optional[ServiceAccount]:

        sa_name = self.sa_name if self.sa_name else get_default_sa_name(self.app_name)
        logger.debug(f"Init ServiceAccount resource: {sa_name}")

        sa_labels = create_component_labels_dict(
            component_name=sa_name,
            app_name=self.app_name,
            labels=self.labels,
        )

        sa_image_pull_secrets: Optional[List[LocalObjectReference]] = None
        if self.image_pull_secrets is not None and isinstance(
            self.image_pull_secrets, list
        ):
            sa_image_pull_secrets = []
            for _ips in self.image_pull_secrets:
                sa_image_pull_secrets.append(LocalObjectReference(name=_ips))

        sa = ServiceAccount(
            api_version=ApiVersion.CORE_V1,
            kind=Kind.SERVICEACCOUNT,
            metadata=ObjectMeta(
                name=sa_name,
                namespace=self.namespace,
                labels=sa_labels,
            ),
            automount_service_account_token=self.automount_service_account_token,
            image_pull_secrets=sa_image_pull_secrets,
            secrets=self.secrets,
        )
        return sa



from pathlib import Path
from typing import Optional, Any, List, Dict
from typing_extensions import Literal

from pydantic import BaseModel

from phidata.infra.aws.api_client import AwsApiClient
from phidata.infra.aws.resource.base import AwsResource
from phidata.utils.cli_console import print_info, print_error, print_subheading
from phidata.utils.log import logger


class CertificateSummary(BaseModel):
    CertificateArn: str
    DomainName: Optional[str] = None


class AcmCertificate(AwsResource):
    """
    You can use Amazon Web Services Certificate Manager (ACM) to manage SSL/TLS
    certificates for your Amazon Web Services-based websites and applications.

    # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/acm.html
    """

    resource_type = "AcmCertificate"
    service_name = "acm"

    # website base domain name, such as example.com
    name: str
    # Fully qualified domain name (FQDN), such as www.example.com,
    # that you want to secure with an ACM certificate.
    #
    # Use an asterisk (*) to create a wildcard certificate that protects several sites in the same domain.
    # For example, *.example.com protects www.example.com, site.example.com, and images.example.com.

    # The first domain name you enter cannot exceed 64 octets, including periods.
    # Each subsequent Subject Alternative Name (SAN), however, can be up to 253 octets in length.

    # If None, defaults to "*.name"
    domain_name: Optional[str] = None
    # The method you want to use if you are requesting a public certificate to validate that you own or control domain.
    # You can validate with DNS or validate with email .
    # We recommend that you use DNS validation.
    validation_method: Literal["EMAIL", "DNS"] = "DNS"
    # Additional FQDNs to be included in the Subject Alternative Name extension of the ACM certificate.
    # For example, add the name www.example.net to a certificate for which the DomainName field is www.example.com
    # if users can reach your site by using either name. The maximum number of domain names that you can add to an
    # ACM certificate is 100. However, the initial quota is 10 domain names. If you need more than 10 names,
    # you must request a quota increase.
    subject_alternative_names: Optional[List[str]] = None
    # Customer chosen string that can be used to distinguish between calls to RequestCertificate .
    # Idempotency tokens time out after one hour. Therefore, if you call RequestCertificate multiple times with
    # the same idempotency token within one hour, ACM recognizes that you are requesting only one certificate
    # and will issue only one. If you change the idempotency token for each call, ACM recognizes that you are
    # requesting multiple certificates.
    idempotency_token: Optional[str] = None
    # The domain name that you want ACM to use to send you emails so that you can validate domain ownership.
    domain_validation_options: Optional[List[dict]] = None
    options: Optional[dict] = None
    certificate_authority_arn: Optional[str] = None
    tags: Optional[List[dict]] = None

    # If True, stores the certificate summary in the file `certificate_summary_file`
    store_cert_summary: bool = False
    # Path for the certificate_summary_file
    certificate_summary_file: Optional[Path] = None

    wait_for_creation = False

    def _create(self, aws_client: AwsApiClient) -> bool:
        """Requests an ACM certificate for use with other Amazon Web Services.

        Args:
            aws_client: The AwsApiClient for the current cluster
        """
        print_info(f"Creating {self.get_resource_type()}: {self.get_resource_name()}")

        # Step 1: Build ACM configuration
        domain_name = self.domain_name
        if domain_name is None:
            domain_name = self.name
        print_info(f"Requesting AcmCertificate for: {domain_name}")

        # create a dict of args which are not null, otherwise aws type validation fails
        not_null_args: Dict[str, Any] = {}
        if self.subject_alternative_names is not None:
            not_null_args["SubjectAlternativeNames"] = self.subject_alternative_names
            print_info("SANs:")
            for san in self.subject_alternative_names:
                print_info(f"    - {san}")
        if self.idempotency_token is not None:
            not_null_args["IdempotencyToken"] = self.idempotency_token
        if self.domain_validation_options is not None:
            not_null_args["DomainValidationOptions"] = self.domain_validation_options
        if self.options is not None:
            not_null_args["Options"] = self.options
        if self.certificate_authority_arn is not None:
            not_null_args["CertificateAuthorityArn"] = self.certificate_authority_arn
        if self.tags is not None:
            not_null_args["Tags"] = self.tags

        # Step 2: Request AcmCertificate
        service_client = self.get_service_client(aws_client)
        try:
            request_cert_response = service_client.request_certificate(
                DomainName=domain_name,
                ValidationMethod=self.validation_method,
                **not_null_args,
            )
            logger.debug(f"AcmCertificate: {request_cert_response}")

            # Validate AcmCertificate creation
            certificate_arn = request_cert_response.get("CertificateArn", None)
            if certificate_arn is not None:
                print_subheading(f"---- Please Note: Certificate ARN ----")
                print_info(f"{certificate_arn}")
                print_subheading(f"--------\n")
                self.active_resource = request_cert_response
                return True
        except Exception as e:
            print_error(f"{self.get_resource_type()} could not be created.")
            print_error(e)
        return False

    def post_create(self, aws_client: AwsApiClient) -> bool:

        # Wait for AcmCertificate to be validated
        if self.wait_for_creation:
            try:
                print_info(f"Waiting for {self.get_resource_type()} to be created.")
                waiter = self.get_service_client(aws_client).get_waiter(
                    "certificate_validated"
                )
                certificate_arn = self.get_certificate_arn(aws_client)
                waiter.wait(
                    CertificateArn=certificate_arn,
                    WaiterConfig={
                        "Delay": self.waiter_delay,
                        "MaxAttempts": self.waiter_max_attempts,
                    },
                )
            except Exception as e:
                print_error("Waiter failed.")
                print_error(e)

        # Store cert summary if needed
        if self.store_cert_summary:
            if self.certificate_summary_file is None:
                print_error("certificate_summary_file not provided")
                return False

            try:
                read_cert_summary = self._read(aws_client)
                if read_cert_summary is None:
                    print_error("certificate_summary not available")
                    return False

                cert_summary = CertificateSummary(**read_cert_summary)
                if not self.certificate_summary_file.exists():
                    self.certificate_summary_file.parent.mkdir(
                        parents=True, exist_ok=True
                    )
                    self.certificate_summary_file.touch(exist_ok=True)
                self.certificate_summary_file.write_text(cert_summary.json(indent=2))
                print_info(
                    f"Certificate Summary stored at: {str(self.certificate_summary_file)}"
                )
            except Exception as e:
                print_error("Could not writing Certificate Summary to file")
                print_error(e)

        return True

    def _read(self, aws_client: AwsApiClient) -> Optional[Any]:
        """Returns the Certificate ARN

        Args:
            aws_client: The AwsApiClient for the current cluster
        """
        logger.debug(f"Reading {self.get_resource_type()}: {self.get_resource_name()}")

        from botocore.exceptions import ClientError

        service_client = self.get_service_client(aws_client)
        try:
            list_certificate_response = service_client.list_certificates()
            # logger.debug(f"AcmCertificate: {list_certificate_response}")

            current_cert = None
            certificate_summary_list = list_certificate_response.get(
                "CertificateSummaryList", []
            )
            for cert_summary in certificate_summary_list:
                domain = cert_summary.get("DomainName", None)
                if domain is not None and domain == self.name:
                    current_cert = cert_summary
            # logger.debug(f"current_cert: {current_cert}")
            # logger.debug(f"current_cert type: {type(current_cert)}")

            if current_cert is not None:
                logger.debug(f"AcmCertificate found: {self.name}")
                self.active_resource = current_cert
        except ClientError as ce:
            logger.debug(f"ClientError: {ce}")
        except Exception as e:
            print_error(f"Error reading {self.get_resource_type()}.")
            print_error(e)
        return self.active_resource

    def _delete(self, aws_client: AwsApiClient) -> bool:
        """Deletes a certificate and its associated private key.

        Args:
            aws_client: The AwsApiClient for the current cluster
        """
        print_info(f"Deleting {self.get_resource_type()}: {self.get_resource_name()}")

        service_client = self.get_service_client(aws_client)
        self.active_resource = None
        try:
            certificate_arn = self.get_certificate_arn(aws_client)
            if certificate_arn is not None:
                delete_cert_response = service_client.delete_certificate(
                    CertificateArn=certificate_arn,
                )
                logger.debug(f"delete_cert_response: {delete_cert_response}")
                print_info(f"AcmCertificate deleted: {self.name}")
            else:
                print_info("AcmCertificate not found")
            return True
        except Exception as e:
            print_error(e)
        return False

    def get_certificate_arn(self, aws_client: AwsApiClient) -> Optional[str]:
        cert_summary = self._read(aws_client)
        if cert_summary is None:
            return None
        cert_arn = cert_summary.get("CertificateArn", None)
        return cert_arn



from pathlib import Path
from typing import Optional, Dict

from phidata.utils.log import logger


def read_env_from_file(env_file: Path) -> Optional[Dict[str, str]]:
    if env_file is not None and env_file.exists() and env_file.is_file():
        logger.debug(f"Reading {env_file}")

        env_dict: Optional[Dict[str, str]] = None

        # logger.debug(f"name: {env_file.name}")
        # logger.debug(f"name: {type(env_file.name)}")
        if env_file.name in (".env", ".envrc"):
            from dotenv.main import dotenv_values

            dotenv_dict: Dict[str, Optional[str]] = dotenv_values(env_file)
            # logger.debug(f"dotenv_dict: {dotenv_dict}")
            env_dict = {k: v for k, v in dotenv_dict.items() if v is not None}

        elif env_file.suffix in (".yaml", ".yml"):
            import yaml

            env_data_from_file = yaml.safe_load(env_file.read_text())
            if env_data_from_file is not None and isinstance(env_data_from_file, dict):
                env_dict = env_data_from_file
            else:
                logger.error(f"Empty env_file: {env_file}")

        # logger.debug(f"env_dict: {env_dict}")
        return env_dict

    return None



from typing import Optional, Any, Callable, Dict, cast, List, Tuple
from typing_extensions import Protocol

from phidata.task import Task, TaskArgs
from phidata.utils.cli_console import print_info
from phidata.utils.log import logger


PythonTaskType = Callable[..., bool]


class PythonTaskArgs(TaskArgs):
    # To run a task we need to execute a function.
    # This python function is called the entrypoint
    # For example:
    #     def simple_python_function(**kwargs) -> bool:
    # The entrypoint() should accept **kwargs as the parameters
    # and return the run_status: True/False as a bool
    # The type for the entrypoint_function is Callable[..., bool].
    #   The entrypoint_function takes any number of/type of arguments
    #   as key=value pairs and returns True/False
    # https://stackoverflow.com/a/39624147
    # https://docs.python.org/3/library/typing.html#typing.Callable
    entrypoint: PythonTaskType

    # Used for internal operations
    entrypoint_args: Optional[Tuple[Any, ...]] = None
    entrypoint_kwargs: Optional[Dict] = None


class PythonTask(Task):
    """A PythonTask executes a python function"""

    def __init__(
        self,
        name: Optional[str] = None,
        entrypoint: Optional[Callable[..., bool]] = None,
        task_id: Optional[str] = None,
        dag_id: Optional[str] = None,
        entrypoint_args: Optional[Tuple[Any, ...]] = None,
        entrypoint_kwargs: Optional[Dict] = None,
        version: Optional[str] = None,
        enabled: bool = True,
    ) -> None:
        super().__init__()
        self.args: Optional[PythonTaskArgs] = None
        if name is not None and entrypoint is not None:
            try:
                self.args = PythonTaskArgs(
                    name=name,
                    entrypoint=entrypoint,
                    task_id=task_id,
                    dag_id=dag_id,
                    entrypoint_args=entrypoint_args,
                    entrypoint_kwargs=entrypoint_kwargs,
                    version=version,
                    enabled=enabled,
                )
            except Exception as e:
                logger.error(f"Args for {self.__class__.__name__} are not valid")
                raise

    @property
    def entrypoint(self) -> Optional[Callable[..., bool]]:
        return self.args.entrypoint if self.args else None

    ######################################################
    ## Run PythonTask
    ######################################################

    def run_in_local_env(self) -> bool:
        """
        Runs a PythonTask in the local environment where phi wf is called from.

        Returns:
            run_status (bool): True if the run was successful
        """
        # PythonTask not yet initialized
        if self.args is None:
            logger.error("PythonTask not initialized")
            return False

        # Important: Validate that PathContext is available
        # This is passed down by the workflow
        if self.path_context is None:
            logger.error("PathContext not available")
            return False

        dry_run: bool = self.run_context.dry_run if self.run_context else False
        if dry_run:
            logger.info("Dry run")
            return True

        run_status: bool = False
        entrypoint = self.entrypoint
        try:
            if self.args is not None and entrypoint is not None:
                run_status = entrypoint(**self.args.dict())
        except Exception as e:
            logger.error("Could not run PythonTask")
            logger.error(e)
        return run_status

    def run_in_docker_container(
        self,
        active_container: Any,
        docker_env: Optional[Dict[str, str]] = None,
    ) -> bool:
        """
        Runs a PythonTask in a docker container.

        Args:
            active_container: The container to run the workflow in
            docker_env:

        Returns:
            run_status (bool): True if the run was successful

        Notes:
            * This function runs in the local environment where phi wf is called from.
            But executes an `airflow` commands in the docker container to run the workflow
            * For the airflow tasks to be available, they need to be added to the workflow DAG
            using add_airflow_tasks_to_dag()
        """
        # PythonTask not yet initialized
        if self.args is None:
            logger.error("PythonTask not initialized")
            return False

        from pathlib import Path

        from docker.errors import APIError
        from docker.models.containers import Container

        from phidata.infra.docker.utils.container import execute_command

        if active_container is None or not isinstance(active_container, Container):
            logger.error("Invalid Container object")
            return False
        container: Container = active_container

        task_id = self.task_id
        dag_id = self.dag_id
        if task_id is None:
            logger.error("task_id unavailable")
            return False
        if dag_id is None:
            logger.error("dag_id unavailable")
            return False

        run_date: str = self.args.run_date.strftime("%Y-%m-%d")
        dry_run: bool = self.run_context.dry_run if self.run_context else False
        if dry_run:
            logger.debug("Dry run")
        detach: bool = self.run_context.detach if self.run_context else False
        if detach:
            logger.debug("Detach ")
        airflow_cmd_dry_run_str: str = " -n" if dry_run else ""

        workflow_file_path: Optional[Path] = (
            self.path_context.workflow_file if self.path_context else None
        )
        airflow_cmd_subdir_str: str = (
            f" -S {str(workflow_file_path.parent)}" if workflow_file_path else ""
        )

        # logger.debug("task_id: {}".format(task_id))
        # logger.debug("dag_id: {}".format(dag_id))
        # logger.debug("run_date: {}".format(run_date))
        # logger.debug("dry_run: {}".format(dry_run))

        run_status: bool = False

        test_wf_cmd = f"airflow tasks test {dag_id} {task_id} {run_date}{airflow_cmd_subdir_str}{airflow_cmd_dry_run_str}"
        dry_run_cmd = f"python {workflow_file_path}"
        command_to_run = dry_run_cmd if dry_run else test_wf_cmd
        try:
            print_info(
                "Command: {}\nContainer: {}".format(command_to_run, container.name)
            )
            run_status = execute_command(
                cmd=command_to_run,
                container=container,
                detach=detach,
                docker_env=docker_env,
            )
        except APIError as e:
            logger.exception(e)
            raise

        return run_status

    def run_in_k8s_container(
        self,
        pod: Any,
        k8s_api_client: Any,
        container_name: Optional[str] = None,
        k8s_env: Optional[Dict[str, str]] = None,
    ) -> bool:
        """
        Runs a PythonTask in a k8s container.

        Args:
            pod: The pod to run the workflow in
            k8s_api_client: K8sApiClient
            container_name: The container to run the workflow in
            k8s_env:

        Returns:
            run_status (bool): True if the run was successful

        Notes:
            * This function runs in the local environment where phi wf is called from.
            But executes an `airflow` commands in the k8s container to run the workflow
            * For the airflow tasks to be available, they need to be added to the workflow DAG
            using add_airflow_tasks_to_dag()
        """
        # PythonTask not yet initialized
        if self.args is None:
            logger.error("PythonTask not initialized")
            return False

        from pathlib import Path

        from phidata.infra.k8s.resource.core.v1.pod import Pod
        from phidata.infra.k8s.api_client import K8sApiClient

        from phidata.infra.k8s.utils.pod import execute_command

        if pod is None or not isinstance(pod, Pod):
            logger.error("Invalid Pod")
            return False
        if k8s_api_client is None or not isinstance(k8s_api_client, K8sApiClient):
            logger.error("Invalid K8sApiClient")
            return False

        task_id = self.task_id
        dag_id = self.dag_id
        if task_id is None:
            logger.error("task_id unavailable")
            return False
        if dag_id is None:
            logger.error("dag_id unavailable")
            return False

        run_date: str = self.args.run_date.strftime("%Y-%m-%d")
        dry_run: Optional[bool] = self.run_context.dry_run if self.run_context else None
        airflow_cmd_dry_run_str: str = " -n" if dry_run else ""

        workflow_file_path: Optional[Path] = (
            self.path_context.workflow_file if self.path_context else None
        )
        airflow_cmd_subdir_str: str = (
            f"-S {str(workflow_file_path.parent)}" if workflow_file_path else ""
        )

        # logger.debug("task_id: {}".format(task_id))
        # logger.debug("dag_id: {}".format(dag_id))
        # logger.debug("run_date: {}".format(run_date))
        # logger.debug("dry_run: {}".format(dry_run))

        run_status: bool = False

        test_wf_cmd = ["airflow", "tasks", "test", dag_id, task_id, run_date]
        dry_run_cmd = ["python", workflow_file_path]
        command_to_run = dry_run_cmd if dry_run else test_wf_cmd
        try:
            print_info("Running command: {}\nContainer: {}".format(command_to_run, pod))
            run_status = execute_command(
                cmd=test_wf_cmd,
                pod=pod,
                k8s_api_client=k8s_api_client,
                container_name=container_name,
                k8s_env=k8s_env,
            )
        except Exception as e:
            logger.error(e)
            raise

        return run_status

    ######################################################
    ## Airflow functions
    ######################################################

    def add_airflow_task(
        self, dag: Any, task_group: Optional[Any] = None
    ) -> Optional[Any]:
        """
        This function adds the PythonTask to a dag.
        Args:
            dag:
            task_group:

        Returns:
            PythonOperator
        """
        # PythonTask not yet initialized
        # logger.info(f"args: {self.args}")
        # logger.info(f"dag: {dag}")
        if self.args is None:
            return None

        # Important: Validate that PathContext is available
        # logger.info(f"path_context: {self.path_context}")
        if self.path_context is None:
            return False

        from airflow.models.dag import DAG
        from airflow.operators.empty import EmptyOperator
        from airflow.operators.python import PythonOperator

        if dag is None or not isinstance(dag, DAG):
            logger.error("Invalid DAG")
            return False
        if task_group is not None:
            from airflow.utils.task_group import TaskGroup

            if not isinstance(dag, TaskGroup):
                logger.error("Invalid TaskGroup")
                return False

        # Get the task_id for the airflow task
        task_id = self.task_id
        # logger.info(f"Creating airflow task: {task_id}")

        if not self.args.enabled:
            return EmptyOperator(task_id=task_id, dag=dag)

        # Function to run
        entrypoint = (
            self.args.entrypoint
            if self.args and self.args.entrypoint is not None
            else simple_python_function
        )
        entrypoint = cast(PythonTaskType, entrypoint)
        # Airflow task
        airflow_task = PythonOperator(
            task_id=task_id,
            dag=dag,
            python_callable=entrypoint,
            op_kwargs=self.args.dict(),
        )
        # logger.debug(f"Airflow task: {task_id} created")

        return airflow_task


def simple_python_function(**kwargs) -> bool:
    logger.error("Invalid entrypoint function")
    logger.error("KWargs: {}".format(kwargs))
    return True


class PythonTaskDecoratorType(Protocol):
    def __call__(
        self,
        *args,
        name: Optional[str] = None,
        task_id: Optional[str] = None,
        dag_id: Optional[str] = None,
        version: Optional[str] = None,
        enabled: bool = True,
        **kwargs,
    ) -> PythonTask:
        ...



from typing import Optional, Any, Dict
from typing_extensions import Literal

from phidata.infra.aws.api_client import AwsApiClient
from phidata.infra.aws.resource.base import AwsResource
from phidata.utils.cli_console import print_info, print_error, print_warning
from phidata.utils.log import logger


class EbsVolume(AwsResource):
    """
    # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ec2.html#volume
    # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ec2.html#EC2.Client.create_volume
    """

    resource_type = "EbsVolume"
    service_name = "ec2"

    # The unique name to give to your volume.
    name: str
    # The size of the volume, in GiBs. You must specify either a snapshot ID or a volume size.
    # If you specify a snapshot, the default is the snapshot size. You can specify a volume size that is
    # equal to or larger than the snapshot size.
    #
    # The following are the supported volumes sizes for each volume type:
    # gp2 and gp3 : 1-16,384
    # io1 and io2 : 4-16,384
    # st1 and sc1 : 125-16,384
    # standard : 1-1,024
    size: int
    # The Availability Zone in which to create the volume.
    availability_zone: str
    # Indicates whether the volume should be encrypted. The effect of setting the encryption state to
    # true depends on the volume origin (new or from a snapshot), starting encryption state, ownership,
    # and whether encryption by default is enabled.
    # Encrypted Amazon EBS volumes must be attached to instances that support Amazon EBS encryption.
    encrypted: Optional[bool] = None
    # The number of I/O operations per second (IOPS). For gp3 , io1 , and io2 volumes, this represents the
    # number of IOPS that are provisioned for the volume. For gp2 volumes, this represents the baseline
    # performance of the volume and the rate at which the volume accumulates I/O credits for bursting.
    #
    # The following are the supported values for each volume type:
    # gp3 : 3,000-16,000 IOPS
    # io1 : 100-64,000 IOPS
    # io2 : 100-64,000 IOPS
    #
    # This parameter is required for io1 and io2 volumes.
    # The default for gp3 volumes is 3,000 IOPS.
    # This parameter is not supported for gp2 , st1 , sc1 , or standard volumes.
    iops: Optional[int] = None
    # The identifier of the Key Management Service (KMS) KMS key to use for Amazon EBS encryption.
    # If this parameter is not specified, your KMS key for Amazon EBS is used. If KmsKeyId is specified,
    # the encrypted state must be true .
    kms_key_id: Optional[str] = None
    # The Amazon Resource Name (ARN) of the Outpost.
    outpost_arn: Optional[str] = None
    # The snapshot from which to create the volume. You must specify either a snapshot ID or a volume size.
    snapshot_id: Optional[str] = None
    # The volume type. This parameter can be one of the following values:
    #
    # General Purpose SSD: gp2 | gp3
    # Provisioned IOPS SSD: io1 | io2
    # Throughput Optimized HDD: st1
    # Cold HDD: sc1
    # Magnetic: standard
    #
    # Default: gp2
    volume_type: Optional[
        Literal["standard", "io_1", "io_2", "gp_2", "sc_1", "st_1", "gp_3"]
    ] = None
    # Checks whether you have the required permissions for the action, without actually making the request,
    # and provides an error response. If you have the required permissions, the error response is DryRunOperation.
    # Otherwise, it is UnauthorizedOperation .
    dry_run: Optional[bool] = None
    # The tags to apply to the volume during creation.
    tags: Optional[Dict[str, str]] = None
    # The tag to use for volume name
    name_tag: str = "Name"
    # Indicates whether to enable Amazon EBS Multi-Attach. If you enable Multi-Attach, you can attach the volume to
    # up to 16 Instances built on the Nitro System in the same Availability Zone. This parameter is supported with
    # io1 and io2 volumes only.
    multi_attach_enabled: Optional[bool] = None
    # The throughput to provision for a volume, with a maximum of 1,000 MiB/s.
    # This parameter is valid only for gp3 volumes.
    # Valid Range: Minimum value of 125. Maximum value of 1000.
    throughput: Optional[int] = None
    # Unique, case-sensitive identifier that you provide to ensure the idempotency of the request.
    # This field is autopopulated if not provided.
    client_token: Optional[str] = None

    wait_for_creation: bool = False

    volume_id: Optional[str] = None

    def _create(self, aws_client: AwsApiClient) -> bool:
        """Creates the EbsVolume

        Args:
            aws_client: The AwsApiClient for the current volume
        """
        print_info(f"Creating {self.get_resource_type()}: {self.get_resource_name()}")

        # Step 1: Build Volume configuration
        # Add name as a tag because volumes do not have names
        tags = {self.name_tag: self.name}
        if self.tags is not None and isinstance(self.tags, dict):
            tags.update(self.tags)

        # create a dict of args which are not null, otherwise aws type validation fails
        not_null_args: Dict[str, Any] = {}
        if self.encrypted:
            not_null_args["Encrypted"] = self.encrypted
        if self.iops:
            not_null_args["Iops"] = self.iops
        if self.kms_key_id:
            not_null_args["KmsKeyId"] = self.kms_key_id
        if self.outpost_arn:
            not_null_args["OutpostArn"] = self.outpost_arn
        if self.snapshot_id:
            not_null_args["SnapshotId"] = self.snapshot_id
        if self.volume_type:
            not_null_args["VolumeType"] = self.volume_type
        if self.dry_run:
            not_null_args["DryRun"] = self.dry_run
        if tags:
            not_null_args["TagSpecifications"] = [
                {
                    "ResourceType": "volume",
                    "Tags": [{"Key": k, "Value": v} for k, v in tags.items()],
                },
            ]
        if self.multi_attach_enabled:
            not_null_args["MultiAttachEnabled"] = self.multi_attach_enabled
        if self.throughput:
            not_null_args["Throughput"] = self.throughput
        if self.client_token:
            not_null_args["ClientToken"] = self.client_token

        # Step 2: Create Volume
        service_resource = self.get_service_resource(aws_client)
        try:
            created_resource = service_resource.create_volume(
                AvailabilityZone=self.availability_zone,
                Size=self.size,
                **not_null_args,
            )
            logger.debug(f"EbsVolume: {created_resource}")

            # Validate Volume creation
            create_time = created_resource.create_time
            self.volume_id = created_resource.volume_id
            logger.debug(f"create_time: {create_time}")
            logger.debug(f"volume_id: {self.volume_id}")
            if create_time is not None:
                print_info(f"EbsVolume created: {self.name}")
                self.active_resource = created_resource
                return True
        except Exception as e:
            print_error(f"{self.get_resource_type()} could not be created.")
            print_error(e)
        return False

    def post_create(self, aws_client: AwsApiClient) -> bool:

        # Wait for Volume to be created
        if self.wait_for_creation:
            try:
                if self.volume_id is not None:
                    print_info(f"Waiting for {self.get_resource_type()} to be created.")
                    waiter = self.get_service_client(aws_client).get_waiter(
                        "volume_available"
                    )
                    waiter.wait(
                        VolumeIds=[self.volume_id],
                        WaiterConfig={
                            "Delay": self.waiter_delay,
                            "MaxAttempts": self.waiter_max_attempts,
                        },
                    )
                else:
                    print_warning("Skipping waiter, no volume_id found")
            except Exception as e:
                print_error("Waiter failed.")
                print_error(e)
        return True

    def _read(self, aws_client: AwsApiClient) -> Optional[Any]:
        """Returns the EbsVolume

        Args:
            aws_client: The AwsApiClient for the current volume
        """
        logger.debug(f"Reading {self.get_resource_type()}: {self.get_resource_name()}")

        from botocore.exceptions import ClientError

        service_resource = self.get_service_resource(aws_client)
        try:
            volume = None
            for _volume in service_resource.volumes.all():
                _volume_tags = _volume.tags
                # logger.debug(f"Found volume: {_volume}")
                # logger.debug(f"Tags: {_volume_tags}")
                if _volume_tags is not None and isinstance(_volume_tags, list):
                    for _tag in _volume_tags:
                        if _tag["Key"] == self.name_tag and _tag["Value"] == self.name:
                            volume = _volume
                            break
                # found volume, break loop
                if volume is not None:
                    break

            if volume is None:
                logger.debug("No EbsVolume found")
                return None

            volume.load()
            create_time = volume.create_time
            self.volume_id = volume.volume_id
            logger.debug(f"create_time: {create_time}")
            logger.debug(f"volume_id: {self.volume_id}")
            if create_time is not None:
                logger.debug(f"EbsVolume found: {self.name}")
                self.active_resource = volume
        except ClientError as ce:
            logger.debug(f"ClientError: {ce}")
        except Exception as e:
            print_error(f"Error reading {self.get_resource_type()}.")
            print_error(e)
        return self.active_resource

    def _delete(self, aws_client: AwsApiClient) -> bool:
        """Deletes the EbsVolume

        Args:
            aws_client: The AwsApiClient for the current volume
        """
        print_info(f"Deleting {self.get_resource_type()}: {self.get_resource_name()}")

        self.active_resource = None
        try:
            volume = self._read(aws_client)
            logger.debug(f"EbsVolume: {volume}")
            if volume is None:
                logger.warning(f"No {self.get_resource_type()} to delete")
                return True

            # detach the volume from all instances
            for attachment in volume.attachments:
                device = attachment.get("Device", None)
                instance_id = attachment.get("InstanceId", None)
                print_info(
                    f"Detaching volume from device: {device}, instance_id: {instance_id}"
                )
                volume.detach_from_instance(
                    Device=device,
                    InstanceId=instance_id,
                )

            # delete volume
            volume.delete()
            print_info(f"EbsVolume deleted: {self.name}")
            return True
        except Exception as e:
            print_error(f"{self.get_resource_type()} could not be deleted.")
            print_error("Please try again or delete resources manually.")
            print_error(e)
        return False

    def get_volume_id(
        self,
        aws_client: Optional[AwsApiClient] = None,
        aws_region: Optional[str] = None,
        aws_profile: Optional[str] = None,
    ) -> Optional[str]:
        aws_api_client = aws_client or AwsApiClient(
            aws_region=aws_region, aws_profile=aws_profile
        )
        if aws_api_client is not None:
            self._read(aws_api_client)
        return self.volume_id



from pathlib import Path
from typing import Optional, List, Dict

from phidata.infra.k8s.args import K8sArgs
from phidata.infra.k8s.enums.manager_status import K8sManagerStatus
from phidata.infra.k8s.exceptions import K8sArgsException
from phidata.infra.k8s.resource.base import K8sResource
from phidata.infra.k8s.resource.group import K8sResourceGroup
from phidata.infra.k8s.worker import K8sWorker
from phidata.utils.log import logger


class K8sManager:
    def __init__(self, k8s_args: K8sArgs):
        if k8s_args is None or not isinstance(k8s_args, K8sArgs):
            raise K8sArgsException("Invalid K8sArgs")
        if k8s_args.workspace_root_path is None or not isinstance(
            k8s_args.workspace_root_path, Path
        ):
            raise K8sArgsException("workspace_root_path invalid")
        if k8s_args.workspace_config_file_path is None or not isinstance(
            k8s_args.workspace_config_file_path, Path
        ):
            raise K8sArgsException("workspace_config_file_path invalid")

        self.k8s_args: K8sArgs = k8s_args
        self.k8s_worker: K8sWorker = K8sWorker(self.k8s_args)
        self.k8s_status: K8sManagerStatus = K8sManagerStatus.PRE_INIT
        logger.debug("**-+-** K8sManager created")

    def get_status(self, refresh: bool = False) -> K8sManagerStatus:
        # logger.debug("Getting K8sManagerStatus")
        if refresh:
            self.k8s_status = K8sManagerStatus.PRE_INIT
            logger.debug(f"K8sManagerStatus: {self.k8s_status.value}")

        if self.k8s_status == K8sManagerStatus.PRE_INIT:
            if self.k8s_worker is not None and self.k8s_worker.is_client_initialized():
                self.k8s_status = K8sManagerStatus.WORKER_READY

        if self.k8s_status == K8sManagerStatus.WORKER_READY:
            if self.k8s_worker is not None and self.k8s_worker.are_resources_active():
                self.k8s_status = K8sManagerStatus.RESOURCES_ACTIVE

        logger.debug(f"K8sManagerStatus: {self.k8s_status.value}")
        return self.k8s_status

    ######################################################
    ## Create Resources
    ######################################################

    def create_resources_dry_run(
        self,
        name_filter: Optional[str] = None,
        type_filter: Optional[str] = None,
        app_filter: Optional[str] = None,
        auto_confirm: Optional[bool] = False,
    ) -> None:

        status = self.get_status()
        if not status.can_create_resources():
            logger.debug("Cannot create resources")
            return
        if self.k8s_worker is None:
            logger.debug("No worker available")
            return

        self.k8s_worker.create_resources_dry_run(
            name_filter=name_filter,
            type_filter=type_filter,
            app_filter=app_filter,
            auto_confirm=auto_confirm,
        )

    def create_resources(
        self,
        name_filter: Optional[str] = None,
        type_filter: Optional[str] = None,
        app_filter: Optional[str] = None,
        auto_confirm: Optional[bool] = False,
    ) -> bool:

        status = self.get_status()
        if not status.can_create_resources():
            logger.debug("Cannot create resources")
            return False
        if self.k8s_worker is None:
            logger.debug("No worker available")
            return False

        return self.k8s_worker.create_resources(
            name_filter=name_filter,
            type_filter=type_filter,
            app_filter=app_filter,
            auto_confirm=auto_confirm,
        )

    def validate_resources_are_created(
        self,
        name_filter: Optional[str] = None,
        type_filter: Optional[str] = None,
        app_filter: Optional[str] = None,
    ) -> bool:

        logger.debug("Validating resources are created...")
        return True

    ######################################################
    ## Get Resources
    ######################################################

    def get_resource_groups(
        self,
        app_filter: Optional[str] = None,
    ) -> Optional[Dict[str, K8sResourceGroup]]:

        status = self.get_status()
        if not status.can_get_resources():
            logger.debug("Cannot get resources")
            return None
        if self.k8s_worker is None:
            logger.debug("No worker available")
            return None

        return self.k8s_worker.build_k8s_resource_groups(app_filter=app_filter)

    def get_resources(
        self,
        name_filter: Optional[str] = None,
        type_filter: Optional[str] = None,
        app_filter: Optional[str] = None,
    ) -> Optional[List[K8sResource]]:

        status = self.get_status()
        if not status.can_get_resources():
            logger.debug("Cannot get resources")
            return None
        if self.k8s_worker is None:
            logger.debug("No worker available")
            return None

        return self.k8s_worker.get_resources(
            name_filter=name_filter, type_filter=type_filter, app_filter=app_filter
        )

    ######################################################
    ## Delete Resources
    ######################################################

    def delete_resources_dry_run(
        self,
        name_filter: Optional[str] = None,
        type_filter: Optional[str] = None,
        app_filter: Optional[str] = None,
        auto_confirm: Optional[bool] = False,
    ) -> None:

        status = self.get_status()
        if not status.can_delete_resources():
            logger.debug("Cannot delete resources")
            return
        if self.k8s_worker is None:
            logger.debug("No worker available")
            return

        # logger.debug("Deleting resources dry run")
        self.k8s_worker.delete_resources_dry_run(
            name_filter=name_filter,
            type_filter=type_filter,
            app_filter=app_filter,
            auto_confirm=auto_confirm,
        )

    def delete_resources(
        self,
        name_filter: Optional[str] = None,
        type_filter: Optional[str] = None,
        app_filter: Optional[str] = None,
        auto_confirm: Optional[bool] = False,
    ) -> bool:

        status = self.get_status()
        if not status.can_delete_resources():
            logger.debug("Cannot delete resources")
            return False
        if self.k8s_worker is None:
            logger.debug("No worker available")
            return False

        return self.k8s_worker.delete_resources(
            name_filter=name_filter,
            type_filter=type_filter,
            app_filter=app_filter,
            auto_confirm=auto_confirm,
        )

    def validate_resources_are_deleted(
        self,
        name_filter: Optional[str] = None,
        type_filter: Optional[str] = None,
        app_filter: Optional[str] = None,
    ) -> bool:

        logger.debug("Validating resources are deleted...")
        return True

    ######################################################
    ## Patch Resources
    ######################################################

    def patch_resources_dry_run(
        self,
        name_filter: Optional[str] = None,
        type_filter: Optional[str] = None,
        app_filter: Optional[str] = None,
        auto_confirm: Optional[bool] = False,
    ) -> None:

        status = self.get_status()
        if not status.can_create_resources():
            logger.debug("Cannot patch resources")
            return
        if self.k8s_worker is None:
            logger.debug("No worker available")
            return

        # logger.debug("Deleting resources dry run")
        self.k8s_worker.patch_resources_dry_run(
            name_filter=name_filter,
            type_filter=type_filter,
            app_filter=app_filter,
            auto_confirm=auto_confirm,
        )

    def patch_resources(
        self,
        name_filter: Optional[str] = None,
        type_filter: Optional[str] = None,
        app_filter: Optional[str] = None,
        auto_confirm: Optional[bool] = False,
    ) -> bool:

        status = self.get_status()
        if not status.can_create_resources():
            logger.debug("Cannot patch resources")
            return False
        if self.k8s_worker is None:
            logger.debug("No worker available")
            return False

        return self.k8s_worker.patch_resources(
            name_filter=name_filter,
            type_filter=type_filter,
            app_filter=app_filter,
            auto_confirm=auto_confirm,
        )

    def validate_resources_are_patched(
        self,
        name_filter: Optional[str] = None,
        type_filter: Optional[str] = None,
        app_filter: Optional[str] = None,
    ) -> bool:

        logger.debug("Validating resources are patched...")
        return True



from textwrap import dedent
from typing import Optional, Any, Dict, List, Literal

from phidata.infra.aws.api_client import AwsApiClient
from phidata.infra.aws.resource.base import AwsResource
from phidata.infra.aws.resource.ecs.container import EcsContainer
from phidata.infra.aws.resource.ecs.volume import EcsVolume
from phidata.infra.aws.resource.iam.role import IamRole
from phidata.utils.cli_console import print_info, print_error
from phidata.utils.log import logger


class EcsTaskDefinition(AwsResource):
    """
    # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ecs.html
    """

    resource_type = "EcsTaskDefinition"
    service_name = "ecs"

    # Name of the task definition.
    # Used as task definition family.
    name: str
    # The family for a task definition.
    # Use name as family if not provided
    # You can use it track multiple versions of the same task definition.
    # The family is used as a name for your task definition.
    family: Optional[str] = None
    # Networking mode to use for the containers in the task.
    # The valid values are none, bridge, awsvpc, and host.
    # If no network mode is specified, the default is bridge.
    network_mode: Optional[Literal["bridge", "host", "awsvpc", "none"]] = None
    # A list of container definitions that describe the different containers that make up the task.
    containers: Optional[List[EcsContainer]] = None
    volumes: Optional[List[EcsVolume]] = None
    placement_constraints: Optional[List[Dict[str, Any]]] = None
    requires_compatibilities: Optional[List[str]] = None
    cpu: Optional[str] = None
    memory: Optional[str] = None
    tags: Optional[List[Dict[str, str]]] = None
    pid_mode: Optional[Literal["host", "task"]] = None
    ipc_mode: Optional[Literal["host", "task", "none"]] = None
    proxy_configuration: Optional[Dict[str, Any]] = None
    inference_accelerators: Optional[List[Dict[str, Any]]] = None
    ephemeral_storage: Optional[Dict[str, Any]] = None
    runtime_platform: Optional[Dict[str, Any]] = None

    # Amazon ECS IAM roles
    # The short name or full Amazon Resource Name (ARN) of the IAM role that containers in this task can assume.
    task_role_arn: Optional[str] = None
    # If task_role_arn is None, a default role is created if create_task_role is True
    create_task_role: bool = False
    # Name for the default role when task_role_arn is None, use "name-task-role" if not provided
    task_role_name: Optional[str] = None
    # Provide a list of policy ARNs to attach to the role
    add_task_role_policy_arns: Optional[List[str]] = None

    # The Amazon Resource Name (ARN) of the task execution role that grants the Amazon ECS container agent permission
    # to make Amazon Web Services API calls on your behalf. The task execution IAM role is required depending on the
    # requirements of your task.
    execution_role_arn: Optional[str] = None
    # If execution_role_arn is None, a default role is created if create_execution_role is True
    create_execution_role: bool = True
    # Name for the default role when execution_role_arn is None, use "name-execution-role" if not provided
    execution_role_name: Optional[str] = None
    # Provide a list of policy ARNs to attach to the role
    add_execution_role_policy_arns: Optional[List[str]] = None

    def get_task_family(self):
        return self.family or self.name

    def _create(self, aws_client: AwsApiClient) -> bool:
        """Create EcsTaskDefinition"""
        print_info(f"Creating {self.get_resource_type()}: {self.get_resource_name()}")

        # Step 1: Get task role arn
        task_role_arn = self.task_role_arn
        if task_role_arn is None and self.create_task_role:
            # Create the IamRole and get task_role_arn
            task_role = self.get_task_role()
            try:
                task_role.create(aws_client)
                task_role_arn = task_role.read(aws_client).arn
                print_info(f"ARN for {task_role.name}: {task_role_arn}")
            except Exception as e:
                print_error("IamRole creation failed, please fix and try again")
                print_error(e)
                return False

        # Step 2: Get execution role arn
        execution_role_arn = self.execution_role_arn
        if execution_role_arn is None and self.create_execution_role:
            # Create the IamRole and get execution_role_arn
            execution_role = self.get_execution_role()
            try:
                execution_role.create(aws_client)
                execution_role_arn = execution_role.read(aws_client).arn
                print_info(f"ARN for {execution_role.name}: {execution_role_arn}")
            except Exception as e:
                print_error("IamRole creation failed, please fix and try again")
                print_error(e)
                return False

        # create a dict of args which are not null, otherwise aws type validation fails
        not_null_args: Dict[str, Any] = {}
        if task_role_arn is not None:
            not_null_args["taskRoleArn"] = task_role_arn
        if execution_role_arn is not None:
            not_null_args["executionRoleArn"] = execution_role_arn
        if self.network_mode is not None:
            not_null_args["networkMode"] = self.network_mode
        if self.containers is not None:
            container_definitions = [
                c.get_container_definition() for c in self.containers
            ]
            not_null_args["containerDefinitions"] = container_definitions
        if self.volumes is not None:
            volume_definitions = [v.get_volume_definition() for v in self.volumes]
            not_null_args["volumes"] = volume_definitions
        if self.placement_constraints is not None:
            not_null_args["placementConstraints"] = self.placement_constraints
        if self.requires_compatibilities is not None:
            not_null_args["requiresCompatibilities"] = self.requires_compatibilities
        if self.cpu is not None:
            not_null_args["cpu"] = self.cpu
        if self.memory is not None:
            not_null_args["memory"] = self.memory
        if self.tags is not None:
            not_null_args["tags"] = self.tags
        if self.pid_mode is not None:
            not_null_args["pidMode"] = self.pid_mode
        if self.ipc_mode is not None:
            not_null_args["ipcMode"] = self.ipc_mode
        if self.proxy_configuration is not None:
            not_null_args["proxyConfiguration"] = self.proxy_configuration
        if self.inference_accelerators is not None:
            not_null_args["inferenceAccelerators"] = self.inference_accelerators
        if self.ephemeral_storage is not None:
            not_null_args["ephemeralStorage"] = self.ephemeral_storage
        if self.runtime_platform is not None:
            not_null_args["runtimePlatform"] = self.runtime_platform

        # Register EcsTaskDefinition
        service_client = self.get_service_client(aws_client)
        try:
            create_response = service_client.register_task_definition(
                family=self.get_task_family(),
                **not_null_args,
            )
            logger.debug(f"EcsTaskDefinition: {create_response}")
            resource_dict = create_response.get("taskDefinition", {})

            # Validate resource creation
            if resource_dict is not None:
                print_info(f"EcsTaskDefinition created: {self.get_resource_name()}")
                self.active_resource = create_response
                return True
        except Exception as e:
            print_error(f"{self.get_resource_type()} could not be created.")
            print_error(e)
        return False

    def _read(self, aws_client: AwsApiClient) -> Optional[Any]:
        """Read EcsTaskDefinition"""
        from botocore.exceptions import ClientError

        logger.debug(f"Reading {self.get_resource_type()}: {self.get_resource_name()}")
        service_client = self.get_service_client(aws_client)
        try:
            describe_response = service_client.describe_task_definition(
                taskDefinition=self.get_task_family()
            )
            logger.debug(f"EcsTaskDefinition: {describe_response}")
            resource = describe_response.get("taskDefinition", None)
            if resource is not None:
                # compare the task definition with the current state
                # if there is a difference, create a new task definition
                # TODO: fix the task_definition_up_to_date function
                # if self.task_definition_up_to_date(task_definition=resource):
                self.active_resource = resource
        except ClientError as ce:
            logger.debug(f"ClientError: {ce}")
        except Exception as e:
            print_error(f"Error reading {self.get_resource_type()}.")
            print_error(e)
        return self.active_resource

    def _delete(self, aws_client: AwsApiClient) -> bool:
        """Delete EcsTaskDefinition"""
        print_info(f"Deleting {self.get_resource_type()}: {self.get_resource_name()}")

        # Step 1: Delete the task role
        if self.task_role_arn is None and self.create_task_role:
            task_role = self.get_task_role()
            try:
                task_role.delete(aws_client)
            except Exception as e:
                print_error(
                    "IamRole deletion failed, please try again or delete manually"
                )
                print_error(e)

        # Step 2: Delete the execution role
        if self.execution_role_arn is None and self.create_execution_role:
            execution_role = self.get_execution_role()
            try:
                execution_role.delete(aws_client)
            except Exception as e:
                print_error(
                    "IamRole deletion failed, please try again or delete manually"
                )
                print_error(e)

        service_client = self.get_service_client(aws_client)
        self.active_resource = None
        try:
            # Get the task definition revisions
            list_response = service_client.list_task_definitions(
                familyPrefix=self.get_task_family(), sort="DESC"
            )
            logger.debug(f"EcsTaskDefinition: {list_response}")
            task_definition_arns = list_response.get("taskDefinitionArns", [])
            if task_definition_arns:
                # Delete all revisions
                for task_definition_arn in task_definition_arns:
                    service_client.deregister_task_definition(
                        taskDefinition=task_definition_arn
                    )
                print_info(f"EcsTaskDefinition deleted: {self.get_resource_name()}")
                return True
        except Exception as e:
            print_error(f"{self.get_resource_type()} could not be deleted.")
            print_error("Please try again or delete resources manually.")
            print_error(e)
        return False

    def _update(self, aws_client: AwsApiClient) -> bool:
        """Update EcsTaskDefinition"""
        print_info(f"Updating {self.get_resource_type()}: {self.get_resource_name()}")

        return self._create(aws_client)

    def get_task_role(self) -> IamRole:
        policy_arns = [
            "arn:aws:iam::aws:policy/service-role/AmazonECSTaskExecutionRolePolicy",
            "arn:aws:iam::aws:policy/CloudWatchFullAccess",
        ]
        if self.add_task_role_policy_arns is not None and isinstance(
            self.add_task_role_policy_arns, list
        ):
            policy_arns.extend(self.add_task_role_policy_arns)

        return IamRole(
            name=self.task_role_name or f"{self.name}-task-role",
            assume_role_policy_document=dedent(
                """\
            {
              "Version": "2012-10-17",
              "Statement": [
                {
                  "Effect": "Allow",
                  "Principal": {
                    "Service": "ecs-tasks.amazonaws.com"
                  },
                  "Action": "sts:AssumeRole"
                }
              ]
            }
            """
            ),
            policy_arns=policy_arns,
        )

    def get_execution_role(self) -> IamRole:
        policy_arns = [
            "arn:aws:iam::aws:policy/service-role/AmazonECSTaskExecutionRolePolicy",
            "arn:aws:iam::aws:policy/CloudWatchFullAccess",
        ]
        if self.add_execution_role_policy_arns is not None and isinstance(
            self.add_execution_role_policy_arns, list
        ):
            policy_arns.extend(self.add_execution_role_policy_arns)

        return IamRole(
            name=self.execution_role_name or f"{self.name}-execution-role",
            assume_role_policy_document=dedent(
                """\
            {
              "Version": "2012-10-17",
              "Statement": [
                {
                  "Effect": "Allow",
                  "Principal": {
                    "Service": "ecs-tasks.amazonaws.com"
                  },
                  "Action": "sts:AssumeRole"
                }
              ]
            }
            """
            ),
            policy_arns=policy_arns,
        )

    def task_definition_up_to_date(self, task_definition: Dict[str, Any]) -> bool:
        """Return True if task_definition from the cluster matches the current state"""

        # Validate container definitions
        if self.containers is not None:
            container_definitions_from_api = task_definition.get("containerDefinitions")
            # Compare the container definitions from the api with the current containers
            # The order of the container definitions should also match
            if container_definitions_from_api is not None and len(
                container_definitions_from_api
            ) == len(self.containers):
                for i, container in enumerate(self.containers):
                    if not container.container_definition_up_to_date(
                        container_definition=container_definitions_from_api[i]
                    ):
                        logger.debug("Container definitions not up to date")
                        return False
            else:
                logger.debug("Container definitions not up to date")
                return False

        # Validate volumes
        if self.volumes is not None:
            volume_definitions_from_api = task_definition.get("volumes")
            # Compare the volume definitions from the api with the current volumes
            # The order of the volume definitions should also match
            if volume_definitions_from_api is not None and len(
                volume_definitions_from_api
            ) == len(self.volumes):
                for i, volume in enumerate(self.volumes):
                    if not volume.volume_definition_up_to_date(
                        volume_definition=volume_definitions_from_api[i]
                    ):
                        logger.debug("Volume definitions not up to date")
                        return False
            else:
                logger.debug("Volume definitions not up to date")
                return False

        # Validate other properties
        if self.task_role_arn is not None:
            if self.task_role_arn != task_definition.get("taskRoleArn"):
                logger.debug(
                    "{} != {}".format(
                        self.task_role_arn, task_definition.get("taskRoleArn")
                    )
                )
                return False
        if self.execution_role_arn is not None:
            if self.execution_role_arn != task_definition.get("executionRoleArn"):
                logger.debug(
                    "{} != {}".format(
                        self.execution_role_arn, task_definition.get("executionRoleArn")
                    )
                )
                return False
        if self.network_mode is not None:
            if self.network_mode != task_definition.get("networkMode"):
                logger.debug(
                    "{} != {}".format(
                        self.network_mode, task_definition.get("networkMode")
                    )
                )
                return False
        if self.placement_constraints is not None:
            if self.placement_constraints != task_definition.get(
                "placementConstraints"
            ):
                logger.debug(
                    "{} != {}".format(
                        self.placement_constraints,
                        task_definition.get("placementConstraints"),
                    )
                )
                return False
        if self.requires_compatibilities is not None:
            if self.requires_compatibilities != task_definition.get(
                "requiresCompatibilities"
            ):
                logger.debug(
                    "{} != {}".format(
                        self.requires_compatibilities,
                        task_definition.get("requiresCompatibilities"),
                    )
                )
                return False
        if self.cpu is not None:
            if self.cpu != task_definition.get("cpu"):
                logger.debug("{} != {}".format(self.cpu, task_definition.get("cpu")))
                return False
        if self.memory is not None:
            if self.memory != task_definition.get("memory"):
                logger.debug(
                    "{} != {}".format(self.memory, task_definition.get("memory"))
                )
                return False
        if self.tags is not None:
            if self.tags != task_definition.get("tags"):
                logger.debug("{} != {}".format(self.tags, task_definition.get("tags")))
                return False
        if self.pid_mode is not None:
            if self.pid_mode != task_definition.get("pidMode"):
                logger.debug(
                    "{} != {}".format(self.pid_mode, task_definition.get("pidMode"))
                )
                return False
        if self.ipc_mode is not None:
            if self.ipc_mode != task_definition.get("ipcMode"):
                logger.debug(
                    "{} != {}".format(self.ipc_mode, task_definition.get("ipcMode"))
                )
                return False
        if self.proxy_configuration is not None:
            if self.proxy_configuration != task_definition.get("proxyConfiguration"):
                logger.debug(
                    "{} != {}".format(
                        self.proxy_configuration,
                        task_definition.get("proxyConfiguration"),
                    )
                )
                return False
        if self.inference_accelerators is not None:
            if self.inference_accelerators != task_definition.get(
                "inferenceAccelerators"
            ):
                logger.debug(
                    "{} != {}".format(
                        self.inference_accelerators,
                        task_definition.get("inferenceAccelerators"),
                    )
                )
                return False
        if self.ephemeral_storage is not None:
            if self.ephemeral_storage != task_definition.get("ephemeralStorage"):
                logger.debug(
                    "{} != {}".format(
                        self.ephemeral_storage, task_definition.get("ephemeralStorage")
                    )
                )
                return False
        if self.runtime_platform is not None:
            if self.runtime_platform != task_definition.get("runtimePlatform"):
                logger.debug(
                    "{} != {}".format(
                        self.runtime_platform, task_definition.get("runtimePlatform")
                    )
                )
                return False

        return True



from typing import List, Optional, Union

from kubernetes.client.models.v1_aws_elastic_block_store_volume_source import (
    V1AWSElasticBlockStoreVolumeSource,
)
from kubernetes.client.models.v1_local_volume_source import V1LocalVolumeSource
from kubernetes.client.models.v1_nfs_volume_source import V1NFSVolumeSource
from kubernetes.client.models.v1_object_reference import V1ObjectReference
from kubernetes.client.models.v1_host_path_volume_source import V1HostPathVolumeSource
from kubernetes.client.models.v1_config_map_volume_source import V1ConfigMapVolumeSource
from kubernetes.client.models.v1_empty_dir_volume_source import V1EmptyDirVolumeSource
from kubernetes.client.models.v1_gce_persistent_disk_volume_source import (
    V1GCEPersistentDiskVolumeSource,
)
from kubernetes.client.models.v1_git_repo_volume_source import V1GitRepoVolumeSource
from kubernetes.client.models.v1_key_to_path import V1KeyToPath
from kubernetes.client.models.v1_persistent_volume_claim_volume_source import (
    V1PersistentVolumeClaimVolumeSource,
)
from kubernetes.client.models.v1_secret_volume_source import V1SecretVolumeSource
from pydantic import Field

from phidata.infra.k8s.resource.base import K8sObject


class KeyToPath(K8sObject):
    resource_type: str = "KeyToPath"

    key: str
    mode: int
    path: str


class AwsElasticBlockStoreVolumeSource(K8sObject):
    """
    # https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_aws_elastic_block_store_volume_source.py
    """

    resource_type: str = "AwsElasticBlockStoreVolumeSource"

    # Filesystem type of the volume that you want to mount.
    # Tip: Ensure that the filesystem type is supported by the host operating system.
    # Examples: "ext4", "xfs", "ntfs". Implicitly inferred to be "ext4" if unspecified.
    # More info: https://kubernetes.io/docs/concepts/storage/volumes#awselasticblockstore
    fs_type: Optional[str] = Field(None, alias="fsType")
    # The partition in the volume that you want to mount. If omitted, the default is to mount
    # by volume name. Examples: For volume /dev/sda1, you specify the partition as "1".
    # Similarly, the volume partition for /dev/sda is "0" (or you can leave the property empty).
    partition: Optional[int] = Field(None, alias="partition")
    # Specify "true" to force and set the ReadOnly property in VolumeMounts to "true".
    # If omitted, the default is "false".
    read_only: Optional[str] = Field(None, alias="readOnly")
    # Unique ID of the persistent disk resource in AWS (Amazon EBS volume).
    # More info: https://kubernetes.io/docs/concepts/storage/volumes#awselasticblockstore
    volume_id: Optional[str] = Field(None, alias="volumeID")

    def get_k8s_object(
        self,
    ) -> V1AWSElasticBlockStoreVolumeSource:
        # Return a V1PersistentVolumeClaimVolumeSource object
        # https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_aws_elastic_block_store_volume_source.py
        _v1_aws_elastic_block_store_volume_source = V1AWSElasticBlockStoreVolumeSource(
            fs_type=self.fs_type,
            partition=self.partition,
            read_only=self.read_only,
            volume_id=self.volume_id,
        )
        return _v1_aws_elastic_block_store_volume_source


class LocalVolumeSource(K8sObject):
    """
    # https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.23/#localvolumesource-v1-core
    """

    resource_type: str = "LocalVolumeSource"

    # The full path to the volume on the node.
    # It can be either a directory or block device (disk, partition, ...).
    path: str
    # Filesystem type to mount. It applies only when the Path is a block device. Must be a filesystem type
    # supported by the host operating system. Ex. "ext4", "xfs", "ntfs".
    # The default value is to auto-select a filesystem if unspecified.
    fs_type: Optional[str] = Field(None, alias="fsType")

    def get_k8s_object(
        self,
    ) -> V1LocalVolumeSource:
        # Return a V1LocalVolumeSource object
        # https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_local_volume_source.py
        _v1_local_volume_source = V1LocalVolumeSource(
            fs_type=self.fs_type,
            path=self.path,
        )
        return _v1_local_volume_source


class HostPathVolumeSource(K8sObject):
    """
    # https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.23/#hostpathvolumesource-v1-core
    """

    resource_type: str = "HostPathVolumeSource"

    # Path of the directory on the host. If the path is a symlink, it will follow the link to the real path.
    # More info: https://kubernetes.io/docs/concepts/storage/volumes#hostpath
    path: str
    # Type for HostPath Volume Defaults to ""
    # More info: https://kubernetes.io/docs/concepts/storage/volumes#hostpath
    type: Optional[str] = None

    def get_k8s_object(
        self,
    ) -> V1HostPathVolumeSource:
        # Return a V1HostPathVolumeSource object
        # https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_host_path_volume_source.py
        _v1_host_path_volume_source = V1HostPathVolumeSource(
            path=self.path,
            type=self.type,
        )
        return _v1_host_path_volume_source


class SecretVolumeSource(K8sObject):
    """
    # https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_secret_volume_source.py
    """

    resource_type: str = "SecretVolumeSource"

    secret_name: str = Field(..., alias="secretName")
    default_mode: Optional[int] = Field(None, alias="defaultMode")
    items: Optional[List[KeyToPath]] = None
    optional: Optional[bool] = None

    def get_k8s_object(self) -> V1SecretVolumeSource:
        # Return a V1SecretVolumeSource object
        # https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_secret_volume_source.py
        _items: Optional[List[V1KeyToPath]] = None
        if self.items:
            _items = []
            for _item in self.items:
                _items.append(
                    V1KeyToPath(
                        key=_item.key,
                        mode=_item.mode,
                        path=_item.path,
                    )
                )

        _v1_secret_volume_source = V1SecretVolumeSource(
            default_mode=self.default_mode,
            items=_items,
            secret_name=self.secret_name,
            optional=self.optional,
        )
        return _v1_secret_volume_source


class ConfigMapVolumeSource(K8sObject):
    """
    # https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_config_map_volume_source.py
    """

    resource_type: str = "ConfigMapVolumeSource"

    name: str
    default_mode: Optional[int] = Field(None, alias="defaultMode")
    items: Optional[List[KeyToPath]] = None
    optional: Optional[bool] = None

    def get_k8s_object(self) -> V1ConfigMapVolumeSource:
        # Return a V1ConfigMapVolumeSource object
        # https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_config_map_volume_source.py
        _items: Optional[List[V1KeyToPath]] = None
        if self.items:
            _items = []
            for _item in self.items:
                _items.append(
                    V1KeyToPath(
                        key=_item.key,
                        mode=_item.mode,
                        path=_item.path,
                    )
                )

        _v1_config_map_volume_source = V1ConfigMapVolumeSource(
            default_mode=self.default_mode,
            items=_items,
            name=self.name,
            optional=self.optional,
        )
        return _v1_config_map_volume_source


class EmptyDirVolumeSource(K8sObject):
    """
    # https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_empty_dir_volume_source.py
    """

    resource_type: str = "EmptyDirVolumeSource"

    medium: Optional[str] = None
    size_limit: Optional[str] = Field(None, alias="sizeLimit")

    def get_k8s_object(self) -> V1EmptyDirVolumeSource:
        # Return a V1EmptyDirVolumeSource object
        # https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_empty_dir_volume_source.py
        _v1_empty_dir_volume_source = V1EmptyDirVolumeSource(
            medium=self.medium,
            size_limit=self.size_limit,
        )
        return _v1_empty_dir_volume_source


class GcePersistentDiskVolumeSource(K8sObject):
    """
    # https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.23/#gcepersistentdiskvolumesource-v1-core
    """

    resource_type: str = "GcePersistentDiskVolumeSource"

    fs_type: str = Field(..., alias="fsType")
    partition: int
    pd_name: str
    read_only: Optional[bool] = Field(None, alias="readOnly")

    def get_k8s_object(
        self,
    ) -> V1GCEPersistentDiskVolumeSource:
        # Return a V1GCEPersistentDiskVolumeSource object
        # https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_gce_persistent_disk_volume_source.py
        _v1_gce_persistent_disk_volume_source = V1GCEPersistentDiskVolumeSource(
            fs_type=self.fs_type,
            partition=self.partition,
            pd_name=self.pd_name,
            read_only=self.read_only,
        )
        return _v1_gce_persistent_disk_volume_source


class GitRepoVolumeSource(K8sObject):
    """
    # https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_git_repo_volume_source.py
    """

    resource_type: str = "GitRepoVolumeSource"

    directory: str
    repository: str
    revision: str

    def get_k8s_object(self) -> V1GitRepoVolumeSource:
        # Return a V1GitRepoVolumeSource object
        # https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_git_repo_volume_source.py
        _v1_git_repo_volume_source = V1GitRepoVolumeSource(
            directory=self.directory,
            repository=self.repository,
            revision=self.revision,
        )
        return _v1_git_repo_volume_source


class PersistentVolumeClaimVolumeSource(K8sObject):
    """
    # https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_persistent_volume_claim_volume_source.py
    """

    resource_type: str = "PersistentVolumeClaimVolumeSource"

    claim_name: str = Field(..., alias="claimName")
    read_only: Optional[bool] = Field(None, alias="readOnly")

    def get_k8s_object(
        self,
    ) -> V1PersistentVolumeClaimVolumeSource:
        # Return a V1PersistentVolumeClaimVolumeSource object
        # https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_persistent_volume_claim_volume_source.py
        _v1_persistent_volume_claim_volume_source = V1PersistentVolumeClaimVolumeSource(
            claim_name=self.claim_name,
            read_only=self.read_only,
        )
        return _v1_persistent_volume_claim_volume_source


class NFSVolumeSource(K8sObject):
    """
    # https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.23/#nfsvolumesource-v1-core
    """

    resource_type: str = "NFSVolumeSource"

    # Path that is exported by the NFS server.
    # More info: https://kubernetes.io/docs/concepts/storage/volumes#nfs
    path: str
    # ReadOnly here will force the NFS export to be mounted with read-only permissions.
    # Defaults to false. More info: https://kubernetes.io/docs/concepts/storage/volumes#nfs
    read_only: Optional[bool] = Field(None, alias="readOnly")
    # Server is the hostname or IP address of the NFS server.
    # More info: https://kubernetes.io/docs/concepts/storage/volumes#nfs
    server: Optional[str] = None

    def get_k8s_object(
        self,
    ) -> V1NFSVolumeSource:
        # Return a V1NFSVolumeSource object
        # https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_nfs_volume_source.py
        _v1_nfs_volume_source = V1NFSVolumeSource(
            path=self.path, read_only=self.read_only, server=self.server
        )
        return _v1_nfs_volume_source


class ClaimRef(K8sObject):
    """
    # https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.23/#persistentvolumespec-v1-core
    """

    resource_type: str = "ClaimRef"

    # Name of the referent.
    # More info: https://kubernetes.io/docs/concepts/overview/working-with-objects/names/#names
    name: Optional[str] = None
    # Namespace of the referent.
    # More info: https://kubernetes.io/docs/concepts/overview/working-with-objects/namespaces/
    namespace: Optional[str] = None

    def get_k8s_object(
        self,
    ) -> V1ObjectReference:
        # Return a V1ObjectReference object
        # https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_object_reference.py
        _v1_object_reference = V1ObjectReference(
            name=self.name,
            namespace=self.namespace,
        )
        return _v1_object_reference


VolumeSourceType = Union[
    AwsElasticBlockStoreVolumeSource,
    ConfigMapVolumeSource,
    EmptyDirVolumeSource,
    GcePersistentDiskVolumeSource,
    GitRepoVolumeSource,
    PersistentVolumeClaimVolumeSource,
    SecretVolumeSource,
    NFSVolumeSource,
]



class AwsConfigException(Exception):
    pass


class AwsManagerException(Exception):
    pass


class AwsWorkerException(Exception):
    pass


class AwsArgsException(Exception):
    pass


class AwsApiClientException(Exception):
    pass


class AwsResourceCreationFailedException(Exception):
    pass



from typing import Dict, List, Optional

from pydantic import BaseModel

from phidata.infra.k8s.enums.api_version import ApiVersion
from phidata.infra.k8s.enums.kind import Kind
from phidata.infra.k8s.resource.rbac_authorization_k8s_io.v1.cluster_role import (
    ClusterRole,
    PolicyRule,
)
from phidata.infra.k8s.create.common.labels import create_component_labels_dict
from phidata.infra.k8s.resource.meta.v1.object_meta import ObjectMeta
from phidata.utils.log import logger


class CreateClusterRole(BaseModel):
    cr_name: str
    app_name: str
    rules: Optional[List[PolicyRule]] = None
    namespace: Optional[str] = None
    labels: Optional[Dict[str, str]] = None

    def create(self) -> Optional[ClusterRole]:
        """Creates the ClusterRole resource"""

        cr_name = self.cr_name
        logger.debug(f"Init ClusterRole resource: {cr_name}")

        cr_labels = create_component_labels_dict(
            component_name=cr_name,
            app_name=self.app_name,
            labels=self.labels,
        )

        cr_rules: List[PolicyRule] = (
            self.rules
            if self.rules
            else [PolicyRule(api_groups=["*"], resources=["*"], verbs=["*"])]
        )

        cr = ClusterRole(
            api_version=ApiVersion.RBAC_AUTH_V1,
            kind=Kind.CLUSTERROLE,
            metadata=ObjectMeta(
                name=cr_name,
                namespace=self.namespace,
                labels=cr_labels,
            ),
            rules=cr_rules,
        )
        return cr



from phidata.utils.enums import ExtendedEnum


class K8sManagerStatus(ExtendedEnum):
    """Enum describing the current status of a K8sManager"""

    # Level 0: The K8sManager has just been created
    PRE_INIT = "PRE_INIT"

    # Level 1: Ready to deploy resources
    WORKER_READY = "WORKER_READY"

    # Level 2: Resources active
    RESOURCES_ACTIVE = "RESOURCES_ACTIVE"

    # Errors
    ERROR = "ERROR"

    def can_create_resources(self) -> bool:
        return self in (
            K8sManagerStatus.WORKER_READY,
            K8sManagerStatus.RESOURCES_ACTIVE,
        )

    def can_delete_resources(self) -> bool:
        return self in (
            K8sManagerStatus.WORKER_READY,
            K8sManagerStatus.RESOURCES_ACTIVE,
        )

    def can_get_resources(self) -> bool:
        return self in (
            K8sManagerStatus.WORKER_READY,
            K8sManagerStatus.RESOURCES_ACTIVE,
        )

    def can_read_resources(self) -> bool:
        return self in (
            K8sManagerStatus.WORKER_READY,
            K8sManagerStatus.RESOURCES_ACTIVE,
        )



from typing import Optional, List

from phidata.infra.base import InfraArgs
from phidata.infra.aws.resource.group import AwsResourceGroup


class AwsArgs(InfraArgs):
    # AwsResourceGroups to deploy
    resources: Optional[List[AwsResourceGroup]] = None
    # Resources dir where aws manifests are stored
    resources_dir: str = "aws"



from typing import List, Optional

from pydantic import BaseModel

from phidata.infra.aws.resource.acm.certificate import AcmCertificate
from phidata.infra.aws.resource.cloudformation.stack import CloudFormationStack
from phidata.infra.aws.resource.ec2.volume import EbsVolume
from phidata.infra.aws.resource.ecs.cluster import EcsCluster
from phidata.infra.aws.resource.ecs.container import EcsContainer
from phidata.infra.aws.resource.ecs.task_definition import EcsTaskDefinition
from phidata.infra.aws.resource.ecs.volume import EcsVolume
from phidata.infra.aws.resource.ecs.service import EcsService
from phidata.infra.aws.resource.eks.cluster import EksCluster
from phidata.infra.aws.resource.eks.fargate_profile import EksFargateProfile
from phidata.infra.aws.resource.eks.node_group import EksNodeGroup
from phidata.infra.aws.resource.eks.kubeconfig import EksKubeconfig
from phidata.infra.aws.resource.iam.role import IamRole
from phidata.infra.aws.resource.iam.policy import IamPolicy
from phidata.infra.aws.resource.glue.crawler import GlueCrawler
from phidata.infra.aws.resource.s3.bucket import S3Bucket
from phidata.infra.aws.resource.emr.cluster import EmrCluster
from phidata.infra.aws.resource.rds.db_cluster import DbCluster
from phidata.infra.aws.resource.rds.db_instance import DbInstance
from phidata.infra.aws.resource.rds.db_subnet_group import DbSubnetGroup
from phidata.infra.aws.resource.elasticache.cluster import CacheCluster


class AwsResourceGroup(BaseModel):
    """The AwsResourceGroup class contains the data for all AwsResources"""

    name: str = "default"
    enabled: bool = True
    weight: int = 100

    iam_roles: Optional[List[IamRole]] = None
    iam_policies: Optional[List[IamPolicy]] = None
    acm_certificates: Optional[List[AcmCertificate]] = None
    s3_buckets: Optional[List[S3Bucket]] = None
    cache_clusters: Optional[List[CacheCluster]] = None
    db_subnet_groups: Optional[List[DbSubnetGroup]] = None
    db_clusters: Optional[List[DbCluster]] = None
    db_instances: Optional[List[DbInstance]] = None
    ecs_clusters: Optional[List[EcsCluster]] = None
    ecs_task_definitions: Optional[List[EcsTaskDefinition]] = None
    ecs_services: Optional[List[EcsService]] = None
    volumes: Optional[List[EbsVolume]] = None
    cloudformation_stacks: Optional[List[CloudFormationStack]] = None
    eks_cluster: Optional[EksCluster] = None
    eks_kubeconfig: Optional[EksKubeconfig] = None
    eks_fargate_profiles: Optional[List[EksFargateProfile]] = None
    eks_nodegroups: Optional[List[EksNodeGroup]] = None
    crawlers: Optional[List[GlueCrawler]] = None
    emr: Optional[List[EmrCluster]] = None



from pathlib import Path
from typing import Optional, List, Dict

from phidata.infra.aws.args import AwsArgs
from phidata.infra.aws.enums import AwsManagerStatus
from phidata.infra.aws.exceptions import AwsArgsException
from phidata.infra.aws.resource.base import AwsResource
from phidata.infra.aws.resource.group import AwsResourceGroup
from phidata.infra.aws.worker import AwsWorker
from phidata.utils.log import logger


class AwsManager:
    def __init__(self, aws_args: AwsArgs):
        if aws_args is None or not isinstance(aws_args, AwsArgs):
            raise AwsArgsException("Invalid AwsArgs")
        if aws_args.workspace_root_path is None or not isinstance(
            aws_args.workspace_root_path, Path
        ):
            raise AwsArgsException("workspace_root_path invalid")
        if aws_args.workspace_config_file_path is None or not isinstance(
            aws_args.workspace_config_file_path, Path
        ):
            raise AwsArgsException("workspace_config_file_path invalid")

        self.aws_args: AwsArgs = aws_args
        self.aws_worker: AwsWorker = AwsWorker(self.aws_args)
        self.aws_status: AwsManagerStatus = AwsManagerStatus.PRE_INIT
        logger.debug("**-+-** AwsManager created")

    def get_status(self, refresh: bool = False) -> AwsManagerStatus:
        # logger.debug("Getting AwsManagerStatus")
        if refresh:
            self.aws_status = AwsManagerStatus.PRE_INIT
            logger.debug(f"AwsManagerStatus: {self.aws_status.value}")

        if self.aws_status == AwsManagerStatus.PRE_INIT:
            if self.aws_worker is not None and self.aws_worker.is_client_initialized():
                self.aws_status = AwsManagerStatus.WORKER_READY

        if self.aws_status == AwsManagerStatus.WORKER_READY:
            if self.aws_worker is not None and self.aws_worker.are_resources_active():
                self.aws_status = AwsManagerStatus.RESOURCES_ACTIVE

        logger.debug(f"AwsManagerStatus: {self.aws_status.value}")
        return self.aws_status

    ######################################################
    ## Create Resources
    ######################################################

    def create_resources_dry_run(
        self,
        name_filter: Optional[str] = None,
        type_filter: Optional[str] = None,
        app_filter: Optional[str] = None,
        auto_confirm: Optional[bool] = False,
    ) -> None:

        status = self.get_status()
        if not status.can_create_resources():
            logger.debug("Cannot create resources")
            return
        if self.aws_worker is None:
            logger.debug("No worker available")
            return

        self.aws_worker.create_resources_dry_run(
            name_filter=name_filter,
            type_filter=type_filter,
            app_filter=app_filter,
            auto_confirm=auto_confirm,
        )

    def create_resources(
        self,
        name_filter: Optional[str] = None,
        type_filter: Optional[str] = None,
        app_filter: Optional[str] = None,
        auto_confirm: Optional[bool] = False,
    ) -> bool:

        status = self.get_status()
        if not status.can_create_resources():
            logger.debug("Cannot create resources")
            return False
        if self.aws_worker is None:
            logger.debug("No worker available")
            return False

        return self.aws_worker.create_resources(
            name_filter=name_filter,
            type_filter=type_filter,
            app_filter=app_filter,
            auto_confirm=auto_confirm,
        )

    def validate_resources_are_created(
        self,
        name_filter: Optional[str] = None,
        type_filter: Optional[str] = None,
        app_filter: Optional[str] = None,
    ) -> bool:

        logger.debug("Validating resources are created...")
        return True

    ######################################################
    ## Get Resources
    ######################################################

    def get_resource_groups(
        self,
        app_filter: Optional[str] = None,
    ) -> Optional[Dict[str, AwsResourceGroup]]:

        status = self.get_status()
        if not status.can_get_resources():
            logger.debug("Cannot get resources")
            return None
        if self.aws_worker is None:
            logger.debug("No worker available")
            return None

        return self.aws_worker.build_aws_resource_groups(app_filter=app_filter)

    def get_resources(
        self,
        name_filter: Optional[str] = None,
        type_filter: Optional[str] = None,
        app_filter: Optional[str] = None,
    ) -> Optional[List[AwsResource]]:

        status = self.get_status()
        if not status.can_get_resources():
            logger.debug("Cannot get AwsResources")
            return None
        if self.aws_worker is None:
            logger.debug("AWSWorker not available")
            return None

        return self.aws_worker.get_resources(
            name_filter=name_filter, type_filter=type_filter, app_filter=app_filter
        )

    ######################################################
    ## Delete Resources
    ######################################################

    def delete_resources_dry_run(
        self,
        name_filter: Optional[str] = None,
        type_filter: Optional[str] = None,
        app_filter: Optional[str] = None,
        auto_confirm: Optional[bool] = False,
    ) -> None:

        status = self.get_status()
        if not status.can_delete_resources():
            logger.debug("Cannot delete resources")
            return
        if self.aws_worker is None:
            logger.debug("No worker available")
            return

        # logger.debug("Deleting resources dry run")
        self.aws_worker.delete_resources_dry_run(
            name_filter=name_filter,
            type_filter=type_filter,
            app_filter=app_filter,
            auto_confirm=auto_confirm,
        )

    def delete_resources(
        self,
        name_filter: Optional[str] = None,
        type_filter: Optional[str] = None,
        app_filter: Optional[str] = None,
        auto_confirm: Optional[bool] = False,
    ) -> bool:

        status = self.get_status()
        if not status.can_delete_resources():
            logger.debug("Cannot delete resources")
            return False
        if self.aws_worker is None:
            logger.debug("No worker available")
            return False

        return self.aws_worker.delete_resources(
            name_filter=name_filter,
            type_filter=type_filter,
            app_filter=app_filter,
            auto_confirm=auto_confirm,
        )

    def validate_resources_are_deleted(
        self,
        name_filter: Optional[str] = None,
        type_filter: Optional[str] = None,
        app_filter: Optional[str] = None,
    ) -> bool:

        logger.debug("Validating resources are deleted...")
        return True

    ######################################################
    ## Patch Resources
    ######################################################

    def patch_resources_dry_run(
        self,
        name_filter: Optional[str] = None,
        type_filter: Optional[str] = None,
        app_filter: Optional[str] = None,
        auto_confirm: Optional[bool] = False,
    ) -> None:

        status = self.get_status()
        if not status.can_create_resources():
            logger.debug("Cannot patch resources")
            return
        if self.aws_worker is None:
            logger.debug("No worker available")
            return

        # logger.debug("Deleting resources dry run")
        self.aws_worker.patch_resources_dry_run(
            name_filter=name_filter,
            type_filter=type_filter,
            app_filter=app_filter,
            auto_confirm=auto_confirm,
        )

    def patch_resources(
        self,
        name_filter: Optional[str] = None,
        type_filter: Optional[str] = None,
        app_filter: Optional[str] = None,
        auto_confirm: Optional[bool] = False,
    ) -> bool:

        status = self.get_status()
        if not status.can_create_resources():
            logger.debug("Cannot patch resources")
            return False
        if self.aws_worker is None:
            logger.debug("No worker available")
            return False

        return self.aws_worker.patch_resources(
            name_filter=name_filter,
            type_filter=type_filter,
            app_filter=app_filter,
            auto_confirm=auto_confirm,
        )

    def validate_resources_are_patched(
        self,
        name_filter: Optional[str] = None,
        type_filter: Optional[str] = None,
        app_filter: Optional[str] = None,
    ) -> bool:

        logger.debug("Validating resources are patched...")
        return True



from collections import OrderedDict
from typing import Dict, List, Type, Union

from phidata.infra.aws.resource.base import AwsResource
from phidata.infra.aws.resource.acm.certificate import AcmCertificate
from phidata.infra.aws.resource.cloudformation.stack import CloudFormationStack
from phidata.infra.aws.resource.ec2.volume import EbsVolume
from phidata.infra.aws.resource.ecs.cluster import EcsCluster
from phidata.infra.aws.resource.ecs.task_definition import EcsTaskDefinition
from phidata.infra.aws.resource.eks.cluster import EksCluster
from phidata.infra.aws.resource.ecs.service import EcsService
from phidata.infra.aws.resource.eks.fargate_profile import EksFargateProfile
from phidata.infra.aws.resource.eks.node_group import EksNodeGroup
from phidata.infra.aws.resource.eks.kubeconfig import EksKubeconfig
from phidata.infra.aws.resource.iam.role import IamRole
from phidata.infra.aws.resource.iam.policy import IamPolicy
from phidata.infra.aws.resource.glue.crawler import GlueCrawler
from phidata.infra.aws.resource.s3.bucket import S3Bucket
from phidata.infra.aws.resource.emr.cluster import EmrCluster
from phidata.infra.aws.resource.rds.db_cluster import DbCluster
from phidata.infra.aws.resource.rds.db_instance import DbInstance
from phidata.infra.aws.resource.rds.db_subnet_group import DbSubnetGroup
from phidata.infra.aws.resource.elasticache.cluster import CacheCluster

# Use this as a type for an object which can hold any Aws Resource
AwsResourceType = Union[
    AcmCertificate,
    CloudFormationStack,
    EbsVolume,
    EksCluster,
    EksFargateProfile,
    EksNodeGroup,
    EksKubeconfig,
    IamRole,
    IamPolicy,
    GlueCrawler,
    S3Bucket,
    DbSubnetGroup,
    DbCluster,
    DbInstance,
    CacheCluster,
    EmrCluster,
    EcsCluster,
    EcsTaskDefinition,
    EcsService,
]

# Use this as an ordered list to iterate over all Aws Resource Classes
# This list is the order in which resources should be installed as well.
AwsResourceTypeList: List[Type[AwsResource]] = [
    IamRole,
    IamPolicy,
    S3Bucket,
    EbsVolume,
    AcmCertificate,
    CloudFormationStack,
    GlueCrawler,
    DbSubnetGroup,
    DbCluster,
    DbInstance,
    CacheCluster,
    EcsCluster,
    EcsTaskDefinition,
    EcsService,
    EksCluster,
    EksKubeconfig,
    EksFargateProfile,
    EksNodeGroup,
    EmrCluster,
]

# Map Aws resource alias' to their type
_aws_resource_type_names: Dict[str, Type[AwsResource]] = {
    aws_type.__name__.lower(): aws_type for aws_type in AwsResourceTypeList
}
_aws_resource_type_aliases: Dict[str, Type[AwsResource]] = {
    "s3": S3Bucket,
    "volume": EbsVolume,
}

AwsResourceAliasToTypeMap: Dict[str, Type[AwsResource]] = dict(
    **_aws_resource_type_names, **_aws_resource_type_aliases
)

# Maps each AwsResource to an install weight
# lower weight AwsResource(s) get installed first
AwsResourceInstallOrder: Dict[str, int] = OrderedDict(
    {
        resource_type.__name__: idx
        for idx, resource_type in enumerate(AwsResourceTypeList, start=1)
    }
)



from phidata.utils.enums import ExtendedEnum


class VolumeType(ExtendedEnum):
    AWS_EBS = "AWS_EBS"
    EMPTY_DIR = "EMPTY_DIR"
    PERSISTENT_VOLUME_CLAIM = "PERSISTENT_VOLUME_CLAIM"
    CONFIG_MAP = "CONFIG_MAP"
    SECRET = "SECRET"
    GCE_PERSISTENT_DISK = "GCE_PERSISTENT_DISK"
    GIT_REPO = "GIT_REPO"
    HOST_PATH = "HOST_PATH"
    LOCAL = "LOCAL"
    NFS = "NFS"



from phidata.utils.enums import ExtendedEnum


class PVAccessMode(ExtendedEnum):
    # the volume can be mounted as read-write by a single node.
    # ReadWriteOnce access mode still can allow multiple pods to access the volume
    # when the pods are running on the same node.
    READ_WRITE_ONCE = "ReadWriteOnce"
    # the volume can be mounted as read-only by many nodes.
    READ_ONLY_MANY = "ReadOnlyMany"
    # the volume can be mounted as read-write by many nodes.
    READ_WRITE_MANY = "ReadWriteMany"
    # the volume can be mounted as read-write by a single Pod. Use ReadWriteOncePod access mode if
    # you want to ensure that only one pod across whole cluster can read that PVC or write to it.
    READ_WRITE_ONCE_POD = "ReadWriteOncePod"



from typing import Any, List, Optional


def compare_dicts(
    dict1: Any,
    dict2: Any,
    ignore_keys: Optional[List[str]] = None,
    compare_keys_only_in_dict1: bool = True,
) -> bool:
    """Compares two dictionaries and returns True if they are equal.

    Args:
        dict1: The first dictionary to compare
        dict2: The second dictionary to compare
        ignore_keys: A list of keys to ignore when comparing the dictionaries
        compare_keys_only_in_dict1: If True, only keys in dict1 will be compared

    Returns:
        True if the dictionaries are equal, False otherwise
    """
    if not (isinstance(dict1, dict) and isinstance(dict2, dict)):
        return False

    # Remove ignored keys from the dictionaries
    if ignore_keys:
        for key in ignore_keys:
            dict1.pop(key, None)
            dict2.pop(key, None)

    # Compare the dictionaries
    if compare_keys_only_in_dict1:
        return dict1 == {k: dict2[k] for k in dict1 if k in dict2}
    else:
        return dict1 == dict2



import json
from typing import List, Optional

from phidata.infra.aws.resource.iam.role import IamRole, IamPolicy


def create_assume_eks_admin_role_policy(
    account_id: str,
    eks_admin_role: IamRole,
    name: Optional[str] = "assume-eks-admin-role.policy",
    version: Optional[str] = "2012-10-17",
    skip_create: bool = False,
    skip_delete: bool = False,
) -> IamPolicy:
    return IamPolicy(
        name=name,
        policy_document=json.dumps(
            {
                "Version": version,
                "Statement": [
                    {
                        "Sid": "VisualEditor0",
                        "Effect": "Allow",
                        "Action": "sts:AssumeRole",
                        "Resource": f"arn:aws:iam::{account_id}:role/{eks_admin_role.name}",
                    }
                ],
            }
        ),
        skip_create=skip_create,
        skip_delete=skip_delete,
    )


def create_eks_admin_policy(
    account_id: str,
    name: Optional[str] = "eks-admin.policy",
    version: Optional[str] = "2012-10-17",
    skip_create: bool = False,
    skip_delete: bool = False,
) -> IamPolicy:
    return IamPolicy(
        name=name,
        policy_document=json.dumps(
            {
                "Version": version,
                "Statement": [
                    {
                        "Sid": "VisualEditor0",
                        "Effect": "Allow",
                        "Action": [
                            "eks:ListClusters",
                            "eks:DescribeAddonVersions",
                            "eks:CreateCluster",
                        ],
                        "Resource": "*",
                    },
                    {
                        "Sid": "VisualEditor1",
                        "Effect": "Allow",
                        "Action": "eks:*",
                        "Resource": f"arn:aws:eks:*:{account_id}:cluster/*",
                    },
                ],
            }
        ),
        skip_create=skip_create,
        skip_delete=skip_delete,
    )


def create_eks_admin_role(
    account_id: str,
    name: Optional[str] = "eks-admin.role",
    version: Optional[str] = "2012-10-17",
    iam_policy_name: Optional[str] = "eks-admin.policy",
    extra_policy_arns: Optional[List[str]] = None,
    skip_create: bool = False,
    skip_delete: bool = False,
) -> IamRole:

    eks_admin_iam_policy = create_eks_admin_policy(
        account_id=account_id,
        name=iam_policy_name,
        version=version,
        skip_create=skip_create,
        skip_delete=skip_delete,
    )
    return IamRole(
        name=name,
        assume_role_policy_document=json.dumps(
            {
                "Version": version,
                "Statement": [
                    {
                        "Effect": "Allow",
                        "Principal": {"AWS": f"arn:aws:iam::{account_id}:root"},
                        "Action": "sts:AssumeRole",
                        "Condition": {},
                    }
                ],
            }
        ),
        policies=[eks_admin_iam_policy],
        policy_arns=extra_policy_arns,
        skip_create=skip_create,
        skip_delete=skip_delete,
    )



from typing import Optional, Any, Dict, List

from phidata.infra.aws.resource.base import AwsResource
from phidata.utils.log import logger


class EcsContainer(AwsResource):
    """
    # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ecs.html
    """

    resource_type = "EcsContainer"
    service_name = "ecs"

    # The name of a container.
    # If you're linking multiple containers together in a task definition, the name of one container can be entered in
    # the links of another container to connect the containers.
    name: str
    # The image used to start a container.
    image: str
    # The private repository authentication credentials to use.
    repository_credentials: Optional[Dict[str, Any]] = None
    # The number of cpu units reserved for the container.
    cpu: Optional[int] = None
    # The amount (in MiB) of memory to present to the container.
    memory: Optional[int] = None
    # The soft limit (in MiB) of memory to reserve for the container.
    memory_reservation: Optional[int] = None
    # The links parameter allows containers to communicate with each other without the need for port mappings.
    links: Optional[List[str]] = None
    # The list of port mappings for the container. Port mappings allow containers to access ports on the host container
    # instance to send or receive traffic.
    port_mappings: Optional[List[Dict[str, Any]]] = None
    # If the essential parameter of a container is marked as true , and that container fails or stops for any reason,
    # all other containers that are part of the task are stopped. If the essential parameter of a container is marked
    # as false , its failure doesn't affect the rest of the containers in a task. If this parameter is omitted,
    # a container is assumed to be essential.
    essential: Optional[bool] = None
    # The entry point that's passed to the container.
    entry_point: Optional[List[str]] = None
    # The command that's passed to the container.
    command: Optional[List[str]] = None
    # The environment variables to pass to a container.
    environment: Optional[List[Dict[str, Any]]] = None
    # The entry point that's passed to the container.
    environment_files: Optional[List[Dict[str, Any]]] = None
    # The mount points for data volumes in your container.
    mount_points: Optional[List[Dict[str, Any]]] = None
    # Data volumes to mount from another container.
    volumes_from: Optional[List[Dict[str, Any]]] = None
    # Linux-specific modifications that are applied to the container, such as Linux kernel capabilities.
    linux_parameters: Optional[Dict[str, Any]] = None
    # The secrets to pass to the container.
    secrets: Optional[List[Dict[str, Any]]] = None
    # The dependencies defined for container startup and shutdown.
    depends_on: Optional[List[Dict[str, Any]]] = None
    # Time duration (in seconds) to wait before giving up on resolving dependencies for a container.
    start_timeout: Optional[int] = None
    # Time duration (in seconds) to wait before the container is forcefully killed if it doesn't exit normally.
    stop_timeout: Optional[int] = None
    # The hostname to use for your container.
    hostname: Optional[str] = None
    # The user to use inside the container.
    user: Optional[str] = None
    # The working directory to run commands inside the container in.
    working_directory: Optional[str] = None
    # When this parameter is true, networking is disabled within the container.
    disable_networking: Optional[bool] = None
    # When this parameter is true, the container is given elevated privileges
    # on the host container instance (similar to the root user).
    privileged: Optional[bool] = None
    readonly_root_filesystem: Optional[bool] = None
    dns_servers: Optional[List[str]] = None
    dns_search_domains: Optional[List[str]] = None
    extra_hosts: Optional[List[Dict[str, Any]]] = None
    docker_security_options: Optional[List[str]] = None
    interactive: Optional[bool] = None
    pseudo_terminal: Optional[bool] = None
    docker_labels: Optional[Dict[str, Any]] = None
    ulimits: Optional[List[Dict[str, Any]]] = None
    log_configuration: Optional[Dict[str, Any]] = None
    health_check: Optional[Dict[str, Any]] = None
    system_controls: Optional[List[Dict[str, Any]]] = None
    resource_requirements: Optional[List[Dict[str, Any]]] = None
    firelens_configuration: Optional[Dict[str, Any]] = None

    def get_container_definition(self) -> Dict[str, Any]:

        container_definition: Dict[str, Any] = {}
        if self.name is not None:
            container_definition["name"] = self.name
        if self.image is not None:
            container_definition["image"] = self.image
        if self.repository_credentials is not None:
            container_definition["repositoryCredentials"] = self.repository_credentials
        if self.cpu is not None:
            container_definition["cpu"] = self.cpu
        if self.memory is not None:
            container_definition["memory"] = self.memory
        if self.memory_reservation is not None:
            container_definition["memoryReservation"] = self.memory_reservation
        if self.links is not None:
            container_definition["links"] = self.links
        if self.port_mappings is not None:
            container_definition["portMappings"] = self.port_mappings
        if self.essential is not None:
            container_definition["essential"] = self.essential
        if self.entry_point is not None:
            container_definition["entryPoint"] = self.entry_point
        if self.command is not None:
            container_definition["command"] = self.command
        if self.environment is not None:
            container_definition["environment"] = self.environment
        if self.environment_files is not None:
            container_definition["environmentFiles"] = self.environment_files
        if self.mount_points is not None:
            container_definition["mountPoints"] = self.mount_points
        if self.volumes_from is not None:
            container_definition["volumesFrom"] = self.volumes_from
        if self.linux_parameters is not None:
            container_definition["linuxParameters"] = self.linux_parameters
        if self.secrets is not None:
            container_definition["secrets"] = self.secrets
        if self.depends_on is not None:
            container_definition["dependsOn"] = self.depends_on
        if self.start_timeout is not None:
            container_definition["startTimeout"] = self.start_timeout
        if self.stop_timeout is not None:
            container_definition["stopTimeout"] = self.stop_timeout
        if self.hostname is not None:
            container_definition["hostname"] = self.hostname
        if self.user is not None:
            container_definition["user"] = self.user
        if self.working_directory is not None:
            container_definition["workingDirectory"] = self.working_directory
        if self.disable_networking is not None:
            container_definition["disableNetworking"] = self.disable_networking
        if self.privileged is not None:
            container_definition["privileged"] = self.privileged
        if self.readonly_root_filesystem is not None:
            container_definition[
                "readonlyRootFilesystem"
            ] = self.readonly_root_filesystem
        if self.dns_servers is not None:
            container_definition["dnsServers"] = self.dns_servers
        if self.dns_search_domains is not None:
            container_definition["dnsSearchDomains"] = self.dns_search_domains
        if self.extra_hosts is not None:
            container_definition["extraHosts"] = self.extra_hosts
        if self.docker_security_options is not None:
            container_definition["dockerSecurityOptions"] = self.docker_security_options
        if self.interactive is not None:
            container_definition["interactive"] = self.interactive
        if self.pseudo_terminal is not None:
            container_definition["pseudoTerminal"] = self.pseudo_terminal
        if self.docker_labels is not None:
            container_definition["dockerLabels"] = self.docker_labels
        if self.ulimits is not None:
            container_definition["ulimits"] = self.ulimits
        if self.log_configuration is not None:
            container_definition["logConfiguration"] = self.log_configuration
        if self.health_check is not None:
            container_definition["healthCheck"] = self.health_check
        if self.system_controls is not None:
            container_definition["systemControls"] = self.system_controls
        if self.resource_requirements is not None:
            container_definition["resourceRequirements"] = self.resource_requirements
        if self.firelens_configuration is not None:
            container_definition["firelensConfiguration"] = self.firelens_configuration

        return container_definition

    def container_definition_up_to_date(
        self, container_definition: Dict[str, Any]
    ) -> bool:
        """This is not working"""

        from phidata.utils.compare import compare_dicts

        if self.name is not None:
            if container_definition.get("name") != self.name:
                logger.debug(
                    "{} != {}".format(self.name, container_definition.get("name"))
                )
                return False
        if self.image is not None:
            if container_definition.get("image") != self.image:
                logger.debug(
                    "{} != {}".format(self.image, container_definition.get("image"))
                )
                return False
        if self.repository_credentials is not None:
            if not compare_dicts(
                self.repository_credentials,
                container_definition.get("repositoryCredentials"),
            ):
                logger.debug(
                    "{} != {}".format(
                        self.repository_credentials,
                        container_definition.get("repositoryCredentials"),
                    )
                )
                return False
        if self.cpu is not None:
            if container_definition.get("cpu") != self.cpu:
                logger.debug(
                    "{} != {}".format(self.cpu, container_definition.get("cpu"))
                )
                return False
        if self.memory is not None:
            if container_definition.get("memory") != self.memory:
                logger.debug(
                    "{} != {}".format(self.memory, container_definition.get("memory"))
                )
                return False
        if self.memory_reservation is not None:
            if container_definition.get("memoryReservation") != self.memory_reservation:
                logger.debug(
                    "{} != {}".format(
                        self.memory_reservation,
                        container_definition.get("memoryReservation"),
                    )
                )
                return False
        if self.links is not None:
            if container_definition.get("links") != self.links:
                logger.debug(
                    "{} != {}".format(self.links, container_definition.get("links"))
                )
                return False
        if self.port_mappings is not None:
            port_mappings_from_api = container_definition.get("portMappings")
            if port_mappings_from_api is not None and len(
                port_mappings_from_api
            ) == len(self.port_mappings):
                for i, port_mapping in enumerate(self.port_mappings):
                    if not compare_dicts(
                        port_mapping,
                        port_mappings_from_api[i],
                    ):
                        logger.debug(
                            "{} != {}".format(
                                port_mapping,
                                port_mappings_from_api[i],
                            )
                        )
                        return False
            else:
                logger.debug("portMappings not up to date")
                return False
        if self.essential is not None:
            if container_definition.get("essential") != self.essential:
                logger.debug(
                    "{} != {}".format(
                        self.essential, container_definition.get("essential")
                    )
                )
                return False
        if self.entry_point is not None:
            if container_definition.get("entryPoint") != self.entry_point:
                logger.debug(
                    "{} != {}".format(
                        self.entry_point, container_definition.get("entryPoint")
                    )
                )
                return False
        if self.command is not None:
            if container_definition.get("command") != self.command:
                logger.debug(
                    "{} != {}".format(self.command, container_definition.get("command"))
                )
                return False
        if self.environment is not None:
            if container_definition.get("environment") != self.environment:
                logger.debug(
                    "{} != {}".format(
                        self.environment, container_definition.get("environment")
                    )
                )
                return False
        if self.environment_files is not None:
            if container_definition.get("environmentFiles") != self.environment_files:
                logger.debug(
                    "{} != {}".format(
                        self.environment_files,
                        container_definition.get("environmentFiles"),
                    )
                )
                return False
        if self.mount_points is not None:
            if container_definition.get("mountPoints") != self.mount_points:
                logger.debug(
                    "{} != {}".format(
                        self.mount_points, container_definition.get("mountPoints")
                    )
                )
                return False
        if self.volumes_from is not None:
            if container_definition.get("volumesFrom") != self.volumes_from:
                logger.debug(
                    "{} != {}".format(
                        self.volumes_from, container_definition.get("volumesFrom")
                    )
                )
                return False
        if self.linux_parameters is not None:
            if container_definition.get("linuxParameters") != self.linux_parameters:
                logger.debug(
                    "{} != {}".format(
                        self.linux_parameters,
                        container_definition.get("linuxParameters"),
                    )
                )
                return False
        if self.secrets is not None:
            if container_definition.get("secrets") != self.secrets:
                logger.debug(
                    "{} != {}".format(self.secrets, container_definition.get("secrets"))
                )
                return False
        if self.depends_on is not None:
            if container_definition.get("dependsOn") != self.depends_on:
                logger.debug(
                    "{} != {}".format(
                        self.depends_on, container_definition.get("dependsOn")
                    )
                )
                return False
        if self.start_timeout is not None:
            if container_definition.get("startTimeout") != self.start_timeout:
                logger.debug(
                    "{} != {}".format(
                        self.start_timeout, container_definition.get("startTimeout")
                    )
                )
                return False
        if self.stop_timeout is not None:
            if container_definition.get("stopTimeout") != self.stop_timeout:
                logger.debug(
                    "{} != {}".format(
                        self.stop_timeout, container_definition.get("stopTimeout")
                    )
                )
                return False
        if self.hostname is not None:
            if container_definition.get("hostname") != self.hostname:
                logger.debug(
                    "{} != {}".format(
                        self.hostname, container_definition.get("hostname")
                    )
                )
                return False
        if self.user is not None:
            if container_definition.get("user") != self.user:
                logger.debug(
                    "{} != {}".format(self.user, container_definition.get("user"))
                )
                return False
        if self.working_directory is not None:
            if container_definition.get("workingDirectory") != self.working_directory:
                logger.debug(
                    "{} != {}".format(
                        self.working_directory,
                        container_definition.get("workingDirectory"),
                    )
                )
                return False
        if self.disable_networking is not None:
            if container_definition.get("disableNetworking") != self.disable_networking:
                logger.debug(
                    "{} != {}".format(
                        self.disable_networking,
                        container_definition.get("disableNetworking"),
                    )
                )
                return False
        if self.privileged is not None:
            if container_definition.get("privileged") != self.privileged:
                logger.debug(
                    "{} != {}".format(
                        self.privileged, container_definition.get("privileged")
                    )
                )
                return False
        if self.readonly_root_filesystem is not None:
            if (
                container_definition.get("readonlyRootFilesystem")
                != self.readonly_root_filesystem
            ):
                logger.debug(
                    "{} != {}".format(
                        self.readonly_root_filesystem,
                        container_definition.get("readonlyRootFilesystem"),
                    )
                )
                return False
        if self.dns_servers is not None:
            if container_definition.get("dnsServers") != self.dns_servers:
                logger.debug(
                    "{} != {}".format(
                        self.dns_servers, container_definition.get("dnsServers")
                    )
                )
                return False
        if self.dns_search_domains is not None:
            if container_definition.get("dnsSearchDomains") != self.dns_search_domains:
                logger.debug(
                    "{} != {}".format(
                        self.dns_search_domains,
                        container_definition.get("dnsSearchDomains"),
                    )
                )
                return False
        if self.extra_hosts is not None:
            if container_definition.get("extraHosts") != self.extra_hosts:
                logger.debug(
                    "{} != {}".format(
                        self.extra_hosts, container_definition.get("extraHosts")
                    )
                )
                return False
        if self.docker_security_options is not None:
            if (
                container_definition.get("dockerSecurityOptions")
                != self.docker_security_options
            ):
                logger.debug(
                    "{} != {}".format(
                        self.docker_security_options,
                        container_definition.get("dockerSecurityOptions"),
                    )
                )
                return False
        if self.interactive is not None:
            if container_definition.get("interactive") != self.interactive:
                logger.debug(
                    "{} != {}".format(
                        self.interactive, container_definition.get("interactive")
                    )
                )
                return False
        if self.pseudo_terminal is not None:
            if container_definition.get("pseudoTerminal") != self.pseudo_terminal:
                logger.debug(
                    "{} != {}".format(
                        self.pseudo_terminal, container_definition.get("pseudoTerminal")
                    )
                )
                return False
        if self.docker_labels is not None:
            if container_definition.get("dockerLabels") != self.docker_labels:
                logger.debug(
                    "{} != {}".format(
                        self.docker_labels, container_definition.get("dockerLabels")
                    )
                )
                return False
        if self.ulimits is not None:
            if container_definition.get("ulimits") != self.ulimits:
                logger.debug(
                    "{} != {}".format(self.ulimits, container_definition.get("ulimits"))
                )
                return False
        if self.log_configuration is not None:
            if container_definition.get("logConfiguration") != self.log_configuration:
                logger.debug(
                    "{} != {}".format(
                        self.log_configuration,
                        container_definition.get("logConfiguration"),
                    )
                )
                return False
        if self.health_check is not None:
            if container_definition.get("healthCheck") != self.health_check:
                logger.debug(
                    "{} != {}".format(
                        self.health_check, container_definition.get("healthCheck")
                    )
                )
                return False
        if self.system_controls is not None:
            if container_definition.get("systemControls") != self.system_controls:
                logger.debug(
                    "{} != {}".format(
                        self.system_controls, container_definition.get("systemControls")
                    )
                )
                return False
        if self.resource_requirements is not None:
            if (
                container_definition.get("resourceRequirements")
                != self.resource_requirements
            ):
                logger.debug(
                    "{} != {}".format(
                        self.resource_requirements,
                        container_definition.get("resourceRequirements"),
                    )
                )
                return False
        if self.firelens_configuration is not None:
            if (
                container_definition.get("firelensConfiguration")
                != self.firelens_configuration
            ):
                logger.debug(
                    "{} != {}".format(
                        self.firelens_configuration,
                        container_definition.get("firelensConfiguration"),
                    )
                )
                return False

        return True
